from boa3 import constants
from boa3.exception import CompilerError
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestPolicyContract(BoaTest):
    default_folder: str = 'test_sc/native_test/policy'

    def test_get_hash(self):
        path = self.get_contract_path('GetHash.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(constants.POLICY_SCRIPT, result)

    def test_get_exec_fee_factor(self):
        path = self.get_contract_path('GetExecFeeFactor.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertIsInstance(result, int)

    def test_get_exec_fee_too_many_parameters(self):
        path = self.get_contract_path('GetExecFeeFactorTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_get_fee_per_byte(self):
        path = self.get_contract_path('GetFeePerByte.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertIsInstance(result, int)

    def test_get_fee_per_byte_too_many_parameters(self):
        path = self.get_contract_path('GetFeePerByteTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_get_storage_price(self):
        path = self.get_contract_path('GetStoragePrice.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertIsInstance(result, int)

    def test_get_storage_price_too_many_parameters(self):
        path = self.get_contract_path('GetStoragePriceTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_is_blocked(self):
        path = self.get_contract_path('IsBlocked.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', bytes(20))
        self.assertEqual(False, result)

    def test_is_blocked_mismatched_type(self):
        path = self.get_contract_path('IsBlockedMismatchedTypeInt.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

        path = self.get_contract_path('IsBlockedMismatchedTypeStr.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

        path = self.get_contract_path('IsBlockedMismatchedTypeBool.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_is_blocked_too_many_parameters(self):
        path = self.get_contract_path('IsBlockedTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_is_blocked_too_few_parameters(self):
        path = self.get_contract_path('IsBlockedTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)



from __future__ import annotations

from typing import Any, Dict

from boa3.model.builtin.interop.nativecontract import PolicyContract
from boa3.model.builtin.native.inativecontractclass import INativeContractClass
from boa3.model.method import Method


class PolicyClass(INativeContractClass):
    """
    A class used to represent Policy native contract
    """

    def __init__(self):
        super().__init__('Policy', PolicyContract)

    @property
    def class_methods(self) -> Dict[str, Method]:
        # avoid recursive import
        from boa3.model.builtin.interop.interop import Interop

        if len(self._class_methods) == 0:
            self._class_methods = {
                'get_fee_per_byte': Interop.GetFeePerByte,
                'get_exec_fee_factor': Interop.GetExecFeeFactor,
                'get_storage_price': Interop.GetStoragePrice,
                'is_blocked': Interop.IsBlocked
            }
        return super().class_methods

    @classmethod
    def build(cls, value: Any = None) -> PolicyClass:
        if value is None or cls._is_type_of(value):
            return _Policy

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, PolicyClass)


_Policy = PolicyClass()



from boa3 import constants
from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.model.builtin.interop.interop import Interop
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3.neo3.contracts import CallFlags
from boa3.neo3.core.types import UInt160, UInt256
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestLedgerContract(BoaTest):
    default_folder: str = 'test_sc/native_test/ledger'

    def test_get_hash(self):
        path = self.get_contract_path('GetHash.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(constants.LEDGER_SCRIPT, result)

    def test_get_block_by_hash(self):
        path = self.get_contract_path('GetBlockByHash.py')
        engine = TestEngine()
        block_index = 1
        engine.increase_block(block_index)
        block_hash = bytes(32)
        result = self.run_smart_contract(engine, path, 'Main', block_hash)
        self.assertIsNone(result)

        current_block = engine.current_block
        self.assertIsNotNone(current_block.hash)

        result = self.run_smart_contract(engine, path, 'Main', current_block.hash)
        self.assertIsInstance(result, list)
        self.assertEqual(10, len(result))
        self.assertEqual(current_block.hash, result[0])
        self.assertEqual(current_block.timestamp, result[4])
        self.assertEqual(current_block.index, result[6])

    def test_get_block_by_index(self):
        path = self.get_contract_path('GetBlockByIndex.py')

        engine = TestEngine()
        index = 0
        result = self.run_smart_contract(engine, path, 'Main', index)
        self.assertIsInstance(result, list)
        self.assertEqual(10, len(result))
        self.assertEqual(index, result[6])

        index = 10
        result = self.run_smart_contract(engine, path, 'Main', index)
        self.assertIsNone(result)

        engine.increase_block(10)
        result = self.run_smart_contract(engine, path, 'Main', index)
        self.assertIsInstance(result, list)
        self.assertEqual(10, len(result))
        self.assertEqual(index, result[6])

    def test_get_block_mismatched_types(self):
        path = self.get_contract_path('GetBlockMismatchedTypes.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_get_transaction(self):
        call_flags = Integer(CallFlags.ALL).to_byte_array(signed=True, min_length=1)
        method = String('getTransaction').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x01'
            + Opcode.LDARG0
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.RET
        )
        path = self.get_contract_path('GetTransaction.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        path_burn_gas = self.get_contract_path('../../interop_test/runtime', 'BurnGas.py')
        engine = TestEngine()

        engine.increase_block()
        sender = bytes(range(20))
        self.run_smart_contract(engine, path_burn_gas, 'main', 1000, signer_accounts=[sender])

        txs = engine.get_transactions()
        self.assertGreater(len(txs), 0)
        hash_ = txs[0].hash
        script = txs[0]._script

        result = self.run_smart_contract(engine, path, 'main', hash_)
        self.assertEqual(8, len(result))
        if isinstance(result[0], str):
            result[0] = String(result[0]).to_bytes()
        self.assertEqual(UInt256(hash_), UInt256(result[0]))  # hash
        self.assertIsInstance(result[1], int)  # version
        self.assertIsInstance(result[2], int)  # nonce
        if isinstance(result[3], str):
            result[3] = String(result[3]).to_bytes()
        self.assertEqual(UInt160(sender), UInt160(result[3]))  # sender
        self.assertIsInstance(result[4], int)  # system_fee
        self.assertIsInstance(result[5], int)  # network_fee
        self.assertIsInstance(result[6], int)  # valid_until_block
        if isinstance(result[7], str):
            result[7] = String(result[7]).to_bytes()
        self.assertEqual(script, result[7])  # script

    def test_get_transaction_mismatched_type(self):
        path = self.get_contract_path('GetTransactionMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_get_transaction_from_block_int(self):
        call_flags = Integer(CallFlags.ALL).to_byte_array(signed=True, min_length=1)
        method = String('getTransactionFromBlock').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x02'
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.RET
        )
        path = self.get_contract_path('GetTransactionFromBlockInt.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        path_burn_gas = self.get_contract_path('../../interop_test/runtime', 'BurnGas.py')
        engine = TestEngine()

        engine.increase_block(10)
        sender = bytes(range(20))
        self.run_smart_contract(engine, path_burn_gas, 'main', 100, signer_accounts=[sender])

        block_10 = engine.current_block
        txs = block_10.get_transactions()
        hash_ = txs[0].hash
        script = txs[0]._script

        engine.increase_block()

        result = self.run_smart_contract(engine, path, 'main', 10, 0)
        self.assertEqual(8, len(result))
        if isinstance(result[0], str):
            result[0] = String(result[0]).to_bytes()
        self.assertEqual(UInt256(hash_), UInt256(result[0]))  # hash
        self.assertIsInstance(result[1], int)  # version
        self.assertIsInstance(result[2], int)  # nonce
        if isinstance(result[3], str):
            result[3] = String(result[3]).to_bytes()
        self.assertEqual(UInt160(sender), UInt160(result[3]))  # sender
        self.assertIsInstance(result[4], int)  # system_fee
        self.assertIsInstance(result[5], int)  # network_fee
        self.assertIsInstance(result[6], int)  # valid_until_block
        if isinstance(result[7], str):
            result[7] = String(result[7]).to_bytes()
        self.assertEqual(script, result[7])  # script

    def test_get_transaction_from_block_uint256(self):
        call_flags = Integer(CallFlags.ALL).to_byte_array(signed=True, min_length=1)
        method = String('getTransactionFromBlock').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x02'
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.RET
        )
        path = self.get_contract_path('GetTransactionFromBlockUInt256.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        path_burn_gas = self.get_contract_path('../../interop_test/runtime', 'BurnGas.py')
        engine = TestEngine()

        engine.increase_block(10)
        sender = bytes(range(20))
        self.run_smart_contract(engine, path_burn_gas, 'main', 100, signer_accounts=[sender])

        block_10 = engine.current_block
        block_hash = block_10.hash
        self.assertIsNotNone(block_hash)
        txs = block_10.get_transactions()
        tx_hash = txs[0].hash
        tx_script = txs[0]._script

        engine.increase_block()

        result = self.run_smart_contract(engine, path, 'main', block_hash, 0)
        self.assertEqual(8, len(result))
        if isinstance(result[0], str):
            result[0] = String(result[0]).to_bytes()
        self.assertEqual(UInt256(tx_hash), UInt256(result[0]))  # hash
        self.assertIsInstance(result[1], int)  # version
        self.assertIsInstance(result[2], int)  # nonce
        if isinstance(result[3], str):
            result[3] = String(result[3]).to_bytes()
        self.assertEqual(UInt160(sender), UInt160(result[3]))  # sender
        self.assertIsInstance(result[4], int)  # system_fee
        self.assertIsInstance(result[5], int)  # network_fee
        self.assertIsInstance(result[6], int)  # valid_until_block
        if isinstance(result[7], str):
            result[7] = String(result[7]).to_bytes()
        self.assertEqual(tx_script, result[7])  # script

    def test_get_transaction_from_block_mismatched_type(self):
        path = self.get_contract_path('GetTransactionFromBlockMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_get_transaction_height(self):
        call_flags = Integer(CallFlags.ALL).to_byte_array(signed=True, min_length=1)
        method = String('getTransactionHeight').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x01'
            + Opcode.LDARG0
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.RET
        )
        path = self.get_contract_path('GetTransactionHeight.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        path_burn_gas = self.get_contract_path('../../interop_test/runtime', 'BurnGas.py')
        engine = TestEngine()

        expected_block_index = 10
        engine.increase_block(expected_block_index)
        self.run_smart_contract(engine, path_burn_gas, 'main', 1000)

        txs = engine.get_transactions()
        self.assertGreater(len(txs), 0)
        hash_ = txs[0].hash

        result = self.run_smart_contract(engine, path, 'main', hash_)
        self.assertEqual(expected_block_index, result)

    def test_get_transaction_height_mismatched_type(self):
        path = self.get_contract_path('GetTransactionHeightMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_get_transaction_signers(self):
        call_flags = Integer(CallFlags.ALL).to_byte_array(signed=True, min_length=1)
        method = String('getTransactionSigners').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x01'
            + Opcode.LDARG0
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.RET
        )
        path = self.get_contract_path('GetTransactionSigners.py')
        output, manifest = self.get_output(path)
        self.assertEqual(expected_output, output)

        path_burn_gas = self.get_contract_path('../../interop_test/runtime', 'BurnGas.py')
        engine = TestEngine()

        example_account_1 = bytes(range(20))
        example_account_2 = bytes(20)
        self.run_smart_contract(engine, path_burn_gas, 'main', 1000,
                                signer_accounts=[example_account_1, example_account_2])

        txs = engine.get_transactions()
        self.assertGreater(len(txs), 0)
        hash_ = txs[0].hash

        result = self.run_smart_contract(engine, path, 'main', hash_)

        self.assertIsInstance(result, list)
        self.assertEqual(len(result), 2)
        self.assertIsInstance(result[0], list)
        self.assertEqual(len(result[0]), len(Interop.SignerType.variables))
        self.assertEqual(result[0][0], String.from_bytes(example_account_1))

    def test_get_transaction_signers_mismatched_type(self):
        path = self.get_contract_path('GetTransactionSignersMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_get_transaction_vm_state(self):
        call_flags = Integer(CallFlags.ALL).to_byte_array(signed=True, min_length=1)
        method = String('getTransactionVMState').to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x01'
            + Opcode.LDARG0
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.RET
        )
        path = self.get_contract_path('GetTransactionVMState.py')
        output, manifest = self.get_output(path)
        self.assertEqual(expected_output, output)

        path_burn_gas = self.get_contract_path('../../interop_test/runtime', 'BurnGas.py')
        engine = TestEngine()

        self.run_smart_contract(engine, path_burn_gas, 'main', 1000)
        expected_vm_state = engine.vm_state.value

        txs = engine.get_transactions()
        self.assertGreater(len(txs), 0)
        hash_ = txs[0].hash
        engine.increase_block()

        result = self.run_smart_contract(engine, path, 'main', hash_)
        self.assertEqual(expected_vm_state, result)

    def test_get_transaction_vm_state_mismatched_type(self):
        path = self.get_contract_path('GetTransactionVMStateMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_get_current_index(self):
        path = self.get_contract_path('GetCurrentIndex.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(engine.height, result)



"""
    Place holder until the VM package gets fully implemented.
"""
from __future__ import annotations

from enum import IntEnum


class VMState(IntEnum):
    NONE = 0
    HALT = 1
    FAULT = 2
    BREAK = 4


def get_vm_state(state_name: str) -> VMState:
    try:
        return VMState[state_name]
    except BaseException:
        return VMState.NONE



__all__ = ['AbortMethod',
           'NeoAccountStateType',
           'Nep11TransferEvent',
           'Nep17TransferEvent',
           'Nep5TransferEvent']

from boa3.model.builtin.contract.abortmethod import AbortMethod
from boa3.model.builtin.contract.neoaccountstatetype import NeoAccountStateType
from boa3.model.builtin.contract.nep11transferevent import Nep11TransferEvent
from boa3.model.builtin.contract.nep17transferevent import Nep17TransferEvent
from boa3.model.builtin.contract.nep5transferevent import Nep5TransferEvent



import os

from boa3 import constants

PROJECT_ROOT_DIRECTORY = constants.PATH_SEPARATOR.join(os.path.dirname(__file__).split(os.sep)[:-1])

# If you didn't install TestEngine in this project's root folder, change this to the path of your .dll folder
TEST_ENGINE_DIRECTORY = f'{PROJECT_ROOT_DIRECTORY}{constants.PATH_SEPARATOR}Neo.TestEngine'



from typing import Dict

from boa3.model.builtin.interop.nativecontract import CryptoLibMethod
from boa3.model.variable import Variable


class VerifyWithECDsaMethod(CryptoLibMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        from boa3.model.type.collection.sequence.ecpointtype import ECPointType
        from boa3.model.type.primitive.bytestringtype import ByteStringType
        from boa3.model.builtin.interop.crypto.namedcurvetype import NamedCurveType

        identifier = 'verify_with_ecdsa'
        native_identifier = 'verifyWithECDsa'
        args: Dict[str, Variable] = {
            'data': Variable(Type.any),
            'pubkey': Variable(ECPointType.build()),
            'signature': Variable(ByteStringType.build()),
            'curve': Variable(NamedCurveType.build())
        }
        super().__init__(identifier, native_identifier, args, return_type=Type.bool)



from boa3 import constants
from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestWhile(BoaTest):
    default_folder: str = 'test_sc/while_test'

    def test_while_constant_condition(self):
        jmpif_address = Integer(6).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.JMP        # begin while
            + jmpif_address
            + Opcode.LDLOC0     # a = a + 2
            + Opcode.PUSH2
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.PUSHF
            + Opcode.JMPIF      # end while False
            + jmp_address
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('ConstantCondition.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(0, result)

    def test_while_variable_condition(self):
        jmpif_address = Integer(12).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-11).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x01'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSH0      # condition = a < value
            + Opcode.LDARG0
            + Opcode.LT
            + Opcode.STLOC1
            + Opcode.JMP        # begin while
            + jmpif_address
            + Opcode.LDLOC0     # a = a + 2
            + Opcode.PUSH2
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.LDLOC0     # condition = a < value * 2
            + Opcode.LDARG0
            + Opcode.PUSH2
            + Opcode.MUL
            + Opcode.LT
            + Opcode.STLOC1
            + Opcode.LDLOC1
            + Opcode.JMPIF      # end while arg0
            + jmp_address
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('VariableCondition.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', -2)
        self.assertEqual(0, result)
        result = self.run_smart_contract(engine, path, 'Main', 5)
        self.assertEqual(10, result)
        result = self.run_smart_contract(engine, path, 'Main', 8)
        self.assertEqual(16, result)

    def test_while_mismatched_type_condition(self):
        path = self.get_contract_path('MismatchedTypeCondition.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_while_no_condition(self):
        path = self.get_contract_path('NoCondition.py')

        with self.assertRaises(SyntaxError):
            output = Boa3.compile(path)

    def test_nested_while(self):
        path = self.get_contract_path('NestedWhile.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(8, result)

    def test_while_else(self):
        jmpif_address = Integer(6).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-5).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x01'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.JMP        # begin while
            + jmpif_address
            + Opcode.LDLOC0     # a = a + 2
            + Opcode.PUSH2
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.PUSHF
            + Opcode.JMPIF      # end while False
            + jmp_address
            + Opcode.LDLOC0     # else
            + Opcode.PUSH1          # a = a + 1
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('WhileElse.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(1, result)

    def test_while_relational_condition(self):
        jmpif_address = Integer(10).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-11).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x00'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSH0      # b = 0
            + Opcode.STLOC1
            + Opcode.JMP        # begin while
            + jmpif_address
            + Opcode.LDLOC0     # a = a + 2
            + Opcode.PUSH2
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.LDLOC1     # b = b + 1
            + Opcode.PUSH1
            + Opcode.ADD
            + Opcode.STLOC1
            + Opcode.LDLOC1
            + Opcode.PUSH10
            + Opcode.LT
            + Opcode.JMPIF      # end while b < 10
            + jmp_address
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('RelationalCondition.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(20, result)

    def test_while_multiple_relational_condition(self):
        jmpif_address = Integer(10).to_byte_array(min_length=1, signed=True)
        jmp_address = Integer(-15).to_byte_array(min_length=1, signed=True)

        expected_output = (
            Opcode.INITSLOT
            + b'\x02'
            + b'\x02'
            + Opcode.PUSH0      # a = 0
            + Opcode.STLOC0
            + Opcode.PUSH0      # b = 0
            + Opcode.STLOC1
            + Opcode.JMP        # begin while
            + jmpif_address
            + Opcode.LDLOC0     # a = a + 2
            + Opcode.PUSH2
            + Opcode.ADD
            + Opcode.STLOC0
            + Opcode.LDLOC1     # b = b + 1
            + Opcode.PUSH1
            + Opcode.ADD
            + Opcode.STLOC1
            + Opcode.LDLOC1
            + Opcode.PUSH10
            + Opcode.LT
            + Opcode.PUSH10
            + Opcode.LDARG1
            + Opcode.LT
            + Opcode.BOOLAND
            + Opcode.JMPIF      # end while b < 10 < arg1
            + jmp_address
            + Opcode.LDLOC0     # return a
            + Opcode.RET
        )

        path = self.get_contract_path('MultipleRelationalCondition.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', '', 1)
        self.assertEqual(0, result)
        result = self.run_smart_contract(engine, path, 'Main', '', 10)
        self.assertEqual(0, result)
        result = self.run_smart_contract(engine, path, 'Main', '', 100)
        self.assertEqual(20, result)

    def test_while_continue(self):
        path = self.get_contract_path('WhileContinue.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(20, result)

    def test_while_break(self):
        path = self.get_contract_path('WhileBreak.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(6, result)

    def test_boa2_while_test(self):
        path = self.get_contract_path('WhileBoa2Test.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(24, result)

    def test_boa2_while_test1(self):
        path = self.get_contract_path('WhileBoa2Test1.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(6, result)

    def test_boa2_while_test2(self):
        path = self.get_contract_path('WhileBoa2Test2.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(6, result)

    def test_while_interop_condition(self):
        path = self.get_contract_path('WhileWithInteropCondition.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'deploy')
        self.assertEqual(True, result)
        contract_hash = engine.executed_script_hash.to_array()
        result = self.run_smart_contract(engine, path, 'test_end_while_jump')
        self.assertEqual(True, result)

        # test notifications inserted into the code for validating if the code flow is correct
        notifications = engine.get_events('notify', origin=contract_hash)
        self.assertEqual(2, len(notifications))

        self.assertEqual(1, len(notifications[0].arguments))
        self.assertIsInstance(notifications[0].arguments[0], list)
        self.assertEqual(4, len(notifications[0].arguments[0]))
        token, executing_script_hash, fee_receiver, fee_amount = notifications[0].arguments[0]
        if isinstance(fee_receiver, str):
            fee_receiver = String(fee_receiver).to_bytes()
        if isinstance(fee_amount, str):
            fee_amount = String(fee_amount).to_bytes()
        if isinstance(fee_amount, bytes):
            fee_amount = Integer.from_bytes(fee_amount)

        self.assertEqual(constants.GAS_SCRIPT, token)
        self.assertEqual(contract_hash, executing_script_hash)
        self.assertEqual(bytes(20), fee_receiver)
        self.assertEqual(10, fee_amount)

        self.assertEqual(1, len(notifications[1].arguments))
        self.assertIsInstance(notifications[1].arguments[0], list)
        self.assertEqual(4, len(notifications[1].arguments[0]))
        token, executing_script_hash, fee_receiver, fee_amount = notifications[1].arguments[0]
        if isinstance(fee_receiver, str):
            fee_receiver = String(fee_receiver).to_bytes()
        if isinstance(fee_amount, str):
            fee_amount = String(fee_amount).to_bytes()
        if isinstance(fee_amount, bytes):
            fee_amount = Integer.from_bytes(fee_amount)

        self.assertEqual(constants.NEO_SCRIPT, token)
        self.assertEqual(contract_hash, executing_script_hash)
        self.assertEqual(bytes(20), fee_receiver)
        self.assertEqual(20, fee_amount)



from typing import List

from boa3.neo.vm.type.String import String
from test_runner.neoxp.command import neoxp
from test_runner.neoxp.command.neoexpresscommand import NeoExpressCommand
from test_runner.neoxp.neoinvoke import NeoInvoke


class NeoExpressBatch:
    def __init__(self):
        self._instructions: List[NeoExpressCommand] = []

    def create_neo_express_checkpoint(self, checkpoint_path: str, overwrite: bool = False):
        self._instructions.append(neoxp.create_checkpoint(checkpoint_path, force=overwrite))

    def deploy_contract(self, nef_path: str):
        self._instructions.append(neoxp.contract.deploy(nef_path, 'genesis'))

    def run_contract(self, invoke: NeoInvoke):
        self._instructions.append(neoxp.contract.run(invoke.contract, invoke.operation, *invoke.args,
                                                     account=invoke.invoker))

    def mint_block(self, block_count: int):
        self._instructions.append(neoxp.fastfwd(block_count))

    def clear(self):
        self._instructions.clear()

    def pop(self):
        return self._instructions.pop()

    def write(self, path: str) -> bool:
        try:
            with open(path, 'wb+') as json_file:
                for instruction in self._instructions:
                    json_file.write(String(f'{instruction.cli_command()}\n').to_bytes())

            return True
        except BaseException:
            return False



from boa3.exception import CompilerError, CompilerWarning
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestNeoTypes(BoaTest):
    default_folder: str = 'test_sc/neo_type_test'

    # region UInt160

    def test_uint160_call_bytes(self):
        path = self.get_contract_path('uint160', 'UInt160CallBytes.py')

        engine = TestEngine()
        value = bytes(20)
        result = self.run_smart_contract(engine, path, 'uint160', value,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        value = bytes(range(20))
        result = self.run_smart_contract(engine, path, 'uint160', value,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'uint160', bytes(10),
                                    expected_result_type=bytes)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'uint160', bytes(30),
                                    expected_result_type=bytes)

    def test_uint160_call_int(self):
        path = self.get_contract_path('uint160', 'UInt160CallInt.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'uint160', 0,
                                         expected_result_type=bytes)
        self.assertEqual(bytes(20), result)

        value = Integer(1_000_000_000).to_byte_array(min_length=20)
        result = self.run_smart_contract(engine, path, 'uint160', 1_000_000_000,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'uint160', -50,
                                    expected_result_type=bytes)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'uint160', bytes(30),
                                    expected_result_type=bytes)

    def test_uint160_call_without_args(self):
        path = self.get_contract_path('uint160', 'UInt160CallWithoutArgs.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'uint160',
                                         expected_result_type=bytes)
        self.assertEqual(bytes(20), result)

    def test_uint160_return_bytes(self):
        path = self.get_contract_path('uint160', 'UInt160ReturnBytes.py')

        engine = TestEngine()
        value = bytes(20)
        result = self.run_smart_contract(engine, path, 'uint160', value,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        value = bytes(range(20))
        result = self.run_smart_contract(engine, path, 'uint160', value,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'uint160', bytes(10),
                                    expected_result_type=bytes)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'uint160', bytes(30),
                                    expected_result_type=bytes)

    def test_uint160_concat_with_bytes(self):
        path = self.get_contract_path('uint160', 'UInt160ConcatWithBytes.py')
        engine = TestEngine()
        value = bytes(20)
        result = self.run_smart_contract(engine, path, 'uint160_method', value,
                                         expected_result_type=bytes)
        self.assertEqual(value + b'123', result)

        value = bytes(range(20))
        result = self.run_smart_contract(engine, path, 'uint160_method', value,
                                         expected_result_type=bytes)
        self.assertEqual(value + b'123', result)

    def test_uint160_mismatched_type(self):
        path = self.get_contract_path('uint160', 'UInt160CallMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region UInt256

    def test_uint256_call_bytes(self):
        path = self.get_contract_path('uint256', 'UInt256CallBytes.py')

        engine = TestEngine()
        value = bytes(32)
        result = self.run_smart_contract(engine, path, 'uint256', value,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        value = bytes(range(32))
        result = self.run_smart_contract(engine, path, 'uint256', value,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'uint256', bytes(20),
                                    expected_result_type=bytes)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'uint256', bytes(30),
                                    expected_result_type=bytes)

    def test_uint256_call_int(self):
        path = self.get_contract_path('uint256', 'UInt256CallInt.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'uint256', 0,
                                         expected_result_type=bytes)
        self.assertEqual(bytes(32), result)

        value = Integer(1_000_000_000).to_byte_array(min_length=32)
        result = self.run_smart_contract(engine, path, 'uint256', 1_000_000_000,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'uint256', -50,
                                    expected_result_type=bytes)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'uint256', bytes(30),
                                    expected_result_type=bytes)

    def test_uint256_call_without_args(self):
        path = self.get_contract_path('uint256', 'UInt256CallWithoutArgs.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'uint256',
                                         expected_result_type=bytes)
        self.assertEqual(bytes(32), result)

    def test_uint256_return_bytes(self):
        path = self.get_contract_path('uint256', 'UInt256ReturnBytes.py')

        engine = TestEngine()
        value = bytes(32)
        result = self.run_smart_contract(engine, path, 'uint256', value,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        value = bytes(range(32))
        result = self.run_smart_contract(engine, path, 'uint256', value,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'uint256', bytes(10),
                                    expected_result_type=bytes)

        with self.assertRaisesRegex(TestExecutionException, self.ASSERT_RESULTED_FALSE_MSG):
            self.run_smart_contract(engine, path, 'uint256', bytes(30),
                                    expected_result_type=bytes)

    def test_uint256_concat_with_bytes(self):
        path = self.get_contract_path('uint256', 'UInt256ConcatWithBytes.py')
        engine = TestEngine()
        value = bytes(32)
        result = self.run_smart_contract(engine, path, 'uint256_method', value,
                                         expected_result_type=bytes)
        self.assertEqual(value + b'123', result)

        value = bytes(range(32))
        result = self.run_smart_contract(engine, path, 'uint256_method', value,
                                         expected_result_type=bytes)
        self.assertEqual(value + b'123', result)

    def test_uint256_mismatched_type(self):
        path = self.get_contract_path('uint256', 'UInt256CallMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    # endregion

    # region ECPoint

    def test_ecpoint_call_bytes(self):
        path = self.get_contract_path('ecpoint', 'ECPointCallBytes.py')
        engine = TestEngine()

        value = bytes(33)
        result = self.run_smart_contract(engine, path, 'ecpoint', value,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        value = bytes(range(33))
        result = self.run_smart_contract(engine, path, 'ecpoint', value,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.UNHANDLED_EXCEPTION_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'ecpoint', bytes(20),
                                    expected_result_type=bytes)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.UNHANDLED_EXCEPTION_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'ecpoint', bytes(30),
                                    expected_result_type=bytes)

    def test_ecpoint_call_without_args(self):
        path = self.get_contract_path('ecpoint', 'ECPointCallWithoutArgs.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_ecpoint_return_bytes(self):
        path = self.get_contract_path('ecpoint', 'ECPointReturnBytes.py')
        engine = TestEngine()

        value = bytes(33)
        result = self.run_smart_contract(engine, path, 'ecpoint', value,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        value = bytes(range(33))
        result = self.run_smart_contract(engine, path, 'ecpoint', value,
                                         expected_result_type=bytes)
        self.assertEqual(value, result)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.UNHANDLED_EXCEPTION_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'ecpoint', bytes(10),
                                    expected_result_type=bytes)

        with self.assertRaisesRegex(TestExecutionException, f'^{self.UNHANDLED_EXCEPTION_MSG_PREFIX}'):
            self.run_smart_contract(engine, path, 'ecpoint', bytes(30),
                                    expected_result_type=bytes)

    def test_ecpoint_concat_with_bytes(self):
        path = self.get_contract_path('ecpoint', 'ECPointConcatWithBytes.py')
        engine = TestEngine()

        value = bytes(33)
        result = self.run_smart_contract(engine, path, 'ecpoint_method', value,
                                         expected_result_type=bytes)
        self.assertEqual(value + b'123', result)

        value = bytes(range(33))
        result = self.run_smart_contract(engine, path, 'ecpoint_method', value,
                                         expected_result_type=bytes)
        self.assertEqual(value + b'123', result)

    def test_ecpoint_mismatched_type(self):
        path = self.get_contract_path('ecpoint', 'ECPointCallMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_ecpoint_script_hash(self):
        path = self.get_contract_path('ecpoint', 'ECPointScriptHash.py')
        engine = TestEngine()

        from boa3.neo import public_key_to_script_hash
        value = bytes(range(33))
        script_hash = public_key_to_script_hash(value)

        result = self.run_smart_contract(engine, path, 'Main', value)
        self.assertEqual(script_hash, result)

    def test_ecpoint_script_hash_from_builtin(self):
        path = self.get_contract_path('ecpoint', 'ECPointScriptHashBuiltinCall.py')
        engine = TestEngine()

        from boa3.neo import public_key_to_script_hash
        value = bytes(range(33))
        script_hash = public_key_to_script_hash(value)

        result = self.run_smart_contract(engine, path, 'Main', value)
        self.assertEqual(script_hash, result)

    # endregion

    # region ByteString

    def test_byte_string_manifest_generation(self):
        path = self.get_contract_path('bytestring', 'ByteStringToBool.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.get_output(path)

        import os
        from boa3.neo.vm.type.AbiType import AbiType

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(1, len(abi['methods']))

        method = abi['methods'][0]

        self.assertIn('parameters', method)
        self.assertEqual(1, len(method['parameters']))
        self.assertIn('type', method['parameters'][0])
        self.assertEqual(AbiType.ByteArray, method['parameters'][0]['type'])

    def test_byte_string_to_bool(self):
        path = self.get_contract_path('bytestring', 'ByteStringToBool.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'to_bool', b'\x00')
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'to_bool', '\x00')
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'to_bool', b'\x01')
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'to_bool', '\x01')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'to_bool', b'\x02')
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'to_bool', '\x02')
        self.assertEqual(True, result)

    def test_byte_string_to_bool_with_builtin(self):
        path = self.get_contract_path('bytestring', 'ByteStringToBoolWithBuiltin.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'to_bool', b'\x00')
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'to_bool', '\x00')
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'to_bool', b'\x01')
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'to_bool', '\x01')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'to_bool', b'\x02')
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'to_bool', '\x02')
        self.assertEqual(True, result)

    def test_byte_string_to_int(self):
        path = self.get_contract_path('bytestring', 'ByteStringToInt.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'to_int', b'\x01\x02')
        self.assertEqual(513, result)
        result = self.run_smart_contract(engine, path, 'to_int', '\x01\x02')
        self.assertEqual(513, result)

    def test_byte_string_to_int_with_builtin(self):
        path = self.get_contract_path('bytestring', 'ByteStringToIntWithBuiltin.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'to_int', b'\x01\x02')
        self.assertEqual(513, result)
        result = self.run_smart_contract(engine, path, 'to_int', '\x01\x02')
        self.assertEqual(513, result)

    def test_byte_string_to_str(self):
        path = self.get_contract_path('bytestring', 'ByteStringToStr.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'to_str', b'abc')
        self.assertEqual('abc', result)

        result = self.run_smart_contract(engine, path, 'to_str', b'123')
        self.assertEqual('123', result)

    def test_byte_string_to_str_with_builtin(self):
        path = self.get_contract_path('bytestring', 'ByteStringToStrWithBuiltin.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'to_str', b'abc')
        self.assertEqual('abc', result)

        result = self.run_smart_contract(engine, path, 'to_str', b'123')
        self.assertEqual('123', result)

    def test_byte_string_to_bytes(self):
        path = self.get_contract_path('bytestring', 'ByteStringToBytes.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'to_bytes', 'abc',
                                         expected_result_type=bytes)
        self.assertEqual(b'abc', result)

        result = self.run_smart_contract(engine, path, 'to_bytes', '123',
                                         expected_result_type=bytes)
        self.assertEqual(b'123', result)

    def test_byte_string_to_bytes_with_builtin(self):
        path = self.get_contract_path('bytestring', 'ByteStringToBytesWithBuiltin.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'to_bytes', 'abc',
                                         expected_result_type=bytes)
        self.assertEqual(b'abc', result)

        result = self.run_smart_contract(engine, path, 'to_bytes', '123',
                                         expected_result_type=bytes)
        self.assertEqual(b'123', result)

    def test_concat_with_bytes(self):
        path = self.get_contract_path('bytestring', 'ConcatWithBytes.py')
        engine = TestEngine()

        prefix = b'12345'
        arg = b'a'
        result = self.run_smart_contract(engine, path, 'concat', arg,
                                         expected_result_type=bytes)
        self.assertEqual(prefix + arg, result)

        arg = '6789'
        result = self.run_smart_contract(engine, path, 'concat', arg,
                                         expected_result_type=bytes)
        self.assertEqual(prefix + String(arg).to_bytes(), result)

    def test_concat_with_str(self):
        path = self.get_contract_path('bytestring', 'ConcatWithStr.py')
        engine = TestEngine()

        prefix = '12345'
        arg = 'a'
        result = self.run_smart_contract(engine, path, 'concat', arg)
        self.assertEqual(prefix + arg, result)

        arg = b'6789'
        result = self.run_smart_contract(engine, path, 'concat', arg)
        self.assertEqual(prefix + String.from_bytes(arg), result)

    def test_concat_with_bytestring(self):
        path = self.get_contract_path('bytestring', 'ConcatWithByteString.py')
        engine = TestEngine()

        prefix = '12345'
        arg = 'a'
        result = self.run_smart_contract(engine, path, 'concat', prefix, arg)
        self.assertEqual(prefix + arg, result)

        arg = b'a'
        result = self.run_smart_contract(engine, path, 'concat', prefix, arg)
        self.assertEqual(prefix + String.from_bytes(arg), result)

        prefix = b'12345'
        arg = b'6789'
        result = self.run_smart_contract(engine, path, 'concat', prefix, arg,
                                         expected_result_type=bytes)
        self.assertEqual(prefix + arg, result)

        arg = '6789'
        result = self.run_smart_contract(engine, path, 'concat', prefix, arg,
                                         expected_result_type=bytes)
        self.assertEqual(prefix + String(arg).to_bytes(), result)

    def test_bytestring_multiplication(self):
        path = self.get_contract_path('bytestring', 'ByteStringMultiplication.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'bytestring_mult', b'a', 4,
                                         expected_result_type=bytes)
        self.assertEqual(b'aaaa', result)
        result = self.run_smart_contract(engine, path, 'bytestring_mult', 'unit', 50)
        self.assertEqual('unit' * 50, result)

    # endregion

    # region Opcode

    def test_opcode_manifest_generation(self):
        path = self.get_contract_path('opcode', 'ConcatWithBytes.py')
        expected_manifest_output = path.replace('.py', '.manifest.json')
        output, manifest = self.get_output(path)

        import os
        from boa3.neo.vm.type.AbiType import AbiType

        self.assertTrue(os.path.exists(expected_manifest_output))
        self.assertIn('abi', manifest)
        abi = manifest['abi']

        self.assertIn('methods', abi)
        self.assertEqual(1, len(abi['methods']))

        method = abi['methods'][0]

        self.assertIn('parameters', method)
        self.assertEqual(1, len(method['parameters']))
        self.assertIn('type', method['parameters'][0])
        self.assertEqual(AbiType.ByteArray, method['parameters'][0]['type'])

    def test_opcode_concat(self):
        from boa3.neo.vm.opcode.Opcode import Opcode
        path = self.get_contract_path('opcode', 'ConcatWithOpcode.py')

        engine = TestEngine()
        expected_result = Opcode.LDARG0 + Opcode.LDARG1 + Opcode.ADD
        result = self.run_smart_contract(engine, path, 'concat')
        self.assertEqual(expected_result, result)

    def test_opcode_concat_with_bytes(self):
        from boa3.neo.vm.opcode.Opcode import Opcode
        path = self.get_contract_path('opcode', 'ConcatWithBytes.py')

        engine = TestEngine()
        concat_bytes = b'12345'
        arg = Opcode.LDARG0
        result = self.run_smart_contract(engine, path, 'concat', arg,
                                         expected_result_type=bytes)
        self.assertEqual(concat_bytes + arg, result)

        arg = Opcode.LDLOC1
        result = self.run_smart_contract(engine, path, 'concat', arg,
                                         expected_result_type=bytes)
        self.assertEqual(concat_bytes + arg, result)

        arg = Opcode.NOP
        result = self.run_smart_contract(engine, path, 'concat', arg,
                                         expected_result_type=bytes)
        self.assertEqual(concat_bytes + arg, result)

    def test_opcode_concat_mismatched_type(self):
        path = self.get_contract_path('opcode', 'ConcatMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_opcode_multiplication(self):
        from boa3.neo.vm.opcode.Opcode import Opcode
        path = self.get_contract_path('opcode', 'OpcodeMultiplication.py')

        engine = TestEngine()

        multiplier = 4
        result = self.run_smart_contract(engine, path, 'opcode_mult', multiplier,
                                         expected_result_type=bytes)
        self.assertEqual(Opcode.NOP * multiplier, result)

        multiplier = 50
        result = self.run_smart_contract(engine, path, 'opcode_mult', multiplier,
                                         expected_result_type=bytes)
        self.assertEqual(Opcode.NOP * multiplier, result)

    # endregion

    # region IsInstance Neo Types

    def test_isinstance_contract(self):
        path = self.get_contract_path('IsInstanceContract.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'is_contract', bytes(10))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_contract', [1, 2, 3])
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_contract', "test_with_string")
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_contract', 42)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'is_get_contract_a_contract')
        self.assertEqual(True, result)

    def test_isinstance_block(self):
        path = self.get_contract_path('IsInstanceBlock.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'is_block', bytes(10))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_block', [1, 2, 3])
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_block', "test_with_string")
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_block', 42)
        self.assertEqual(False, result)

        engine.increase_block(10)
        result = self.run_smart_contract(engine, path, 'get_block_is_block', 5)
        self.assertEqual(True, result)

    def test_transaction_cast_and_get_hash(self):
        path = self.get_contract_path('CastTransactionGetHash.py')
        self.assertCompilerLogs(CompilerWarning.TypeCasting, path)

    def test_transaction_implicit_cast_and_get_hash(self):
        path = self.get_contract_path('ImplicitCastTransactionGetHash.py')
        self.assertCompilerLogs(CompilerWarning.TypeCasting, path)

    def test_transaction_cast_and_assign_hash_to_variable(self):
        path = self.get_contract_path('CastTransactionGetHashToVariable.py')
        self.assertCompilerLogs(CompilerWarning.TypeCasting, path)

    def test_isinstance_transaction(self):
        path = self.get_contract_path('IsInstanceTransaction.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'is_tx', bytes(10))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_tx', [1, 2, 3])
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_tx', "test_with_string")
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_tx', 42)
        self.assertEqual(False, result)

        txs = engine.current_block.get_transactions()
        self.assertGreater(len(txs), 0)
        tx_hash = txs[-1].hash

        result = self.run_smart_contract(engine, path, 'get_transaction_is_tx', tx_hash)
        self.assertEqual(True, result)

    def test_isinstance_notification(self):
        path = self.get_contract_path('IsInstanceNotification.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'is_notification', bytes(10))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_notification', [1, 2, 3])
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_notification', "test_with_string")
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_notification', 42)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'get_notifications_is_notification')
        self.assertEqual(True, result)

    def test_isinstance_storage_context(self):
        path = self.get_contract_path('IsInstanceStorageContext.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'is_context', bytes(10))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_context', [1, 2, 3])
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_context', "test_with_string")
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_context', 42)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'get_context_is_context')
        self.assertEqual(True, result)

    def test_isinstance_storage_map(self):
        path = self.get_contract_path('IsInstanceStorageMap.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'is_storage_map', bytes(10))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_storage_map', [1, 2, 3])
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_storage_map', "test_with_string")
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_storage_map', 42)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'create_map_is_storage_map')
        self.assertEqual(True, result)

    def test_isinstance_iterator(self):
        path = self.get_contract_path('IsInstanceIterator.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'is_iterator', bytes(10))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_iterator', [1, 2, 3])
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_iterator', "test_with_string")
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_iterator', 42)
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'storage_find_is_context')
        self.assertEqual(True, result)

    def test_isinstance_ecpoint(self):
        path = self.get_contract_path('IsInstanceECPoint.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'is_ecpoint', bytes(10))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_ecpoint', bytes(33))
        self.assertEqual(True, result)
        result = self.run_smart_contract(engine, path, 'is_ecpoint', bytes(30))
        self.assertEqual(False, result)
        result = self.run_smart_contract(engine, path, 'is_ecpoint', 42)
        self.assertEqual(False, result)

    # endregion



from typing import List, Dict, Any, Union, Optional, Tuple

from boa3.builtin.type import ByteString, Event


def CreateNewEvent(arguments: List[Tuple[str, type]] = [], event_name: str = '') -> Event:
    """
    Creates a new event.

    :param arguments: the list of the events args' names and types
    :type arguments: List[Tuple[str, type]]
    :param event_name: custom name of the event. It's filled with the variable name if not specified
    :type event_name: str
    :return: the new event
    :rtype: Event
    """
    pass


def public(name: str = None, safe: bool = True, *args, **kwargs):
    """
    This decorator identifies which methods should be included in the abi file.

    :param name: Identifier for this method that'll be used on the abi. If not specified, it'll be the same
     identifier from Python method definition
    :type name: str
    :param safe: Whether this method is an abi safe method. False by default
    :type safe: bool
    """
    def decorator_wrapper(*args, **kwargs):
        pass
    return decorator_wrapper


def metadata(*args):
    """
    This decorator identifies the function that returns the metadata object of the smart contract.
    This can be used to only one function. Using this decorator in multiple functions will raise a compiler error.
    """
    pass


def contract(script_hash: ByteString):
    """
    This decorator identifies a class that should be interpreted as an interface to an existing contract.

    :param script_hash: Script hash of the interfaced contract
    :type script_hash: str or bytes
    """
    def decorator_wrapper(cls, *args, **kwargs):
        cls.hash = script_hash
        return cls
    return decorator_wrapper


def display_name(name: str):
    """
    This decorator identifies which methods from a contract interface should have a different identifier from the one
    interfacing it. It only works in contract interface classes.

    :param name: Method identifier from the contract manifest.
    :type name: str
    """
    pass


class NeoMetadata:
    """
    This class stores the smart contract manifest information.

    :ivar name: the smart contract name. Will be the name of the file by default;
    :vartype type name: str
    :ivar supported_standards: Neo standards supported by this smart contract. Empty by default;
    :vartype supported_standards: List[str]
    :ivar permissions: a list of contracts and methods that this smart contract permits to invoke and call. All
     contracts and methods permitted by default;
    :vartype permissions: List[str]
    :ivar trusts: a list of contracts that this smart contract trust. Empty by default;
    :vartype trusts: List[str]
    :ivar author: the smart contract author. None by default;
    :vartype author: str or None
    :ivar email: the smart contract author email. None by default;
    :vartype email: str or None
    :ivar description: the smart contract description. None by default;
    :vartype description: str or None
    """
    from boa3.constants import IMPORT_WILDCARD

    def __init__(self):
        self.name: str = ''
        self.source: Optional[str] = None
        self.supported_standards: List[str] = []
        self._trusts: List[str] = []
        self._permissions: List[dict] = []
        self._groups: List[dict] = []

    @property
    def extras(self) -> Dict[str, Any]:
        """
        Gets the metadata extra information.

        :return: a dictionary that maps each extra value with its name. Empty by default
        """
        # list the variables names that are part of the manifest
        specific_field_names = ['name',
                                'source',
                                'supported_standards',
                                ]
        extra = {}

        for var_name, var_value in vars(self).items():
            if var_name in specific_field_names:
                continue

            if var_value is not None and not var_name.startswith('_'):
                extra_field = var_name.title().replace('_', '')
                extra[extra_field] = var_value

        return extra

    @property
    def trusts(self) -> List[str]:
        return self._trusts.copy()

    @property
    def permissions(self) -> List[dict]:
        return self._permissions.copy()

    @property
    def groups(self) -> List[dict]:
        return self._groups.copy()

    def add_trusted_source(self, hash_or_address: str):
        """
        Adds a valid contract hash, valid group public key, or the '*' wildcard to trusts.

        :param hash_or_address: a contract hash, group public key or '*'
        :type hash_or_address: str
        """
        if not isinstance(hash_or_address, str):
            return

        from boa3.constants import IMPORT_WILDCARD
        if self._trusts == [IMPORT_WILDCARD]:
            return

        if hash_or_address == IMPORT_WILDCARD:
            self._trusts = [IMPORT_WILDCARD]

        # verifies if it's a valid contract hash
        elif self._verify_is_valid_contract_hash(hash_or_address):
            if hash_or_address not in self._trusts:
                self._trusts.append(hash_or_address.lower())

        # verifies if it's a valid public key
        elif self._verify_is_valid_public_key(hash_or_address):
            if hash_or_address not in self._trusts:
                self._trusts.append(hash_or_address.lower())

    def add_group(self, pub_key: str, signature: str):
        """
        Adds a pair of public key and signature to the groups in the manifest.

        :param pub_key: public key of the group
        :type pub_key: str
        :param signature: signature of the contract hash encoded in Base64
        :type signature: str
        """

        if not isinstance(pub_key, str) or not isinstance(signature, str):
            return

        # verifies if pub_key is a valid value
        if not self._verify_is_valid_public_key(pub_key):
            return

        new_group = {
            'pubkey': pub_key.lower(),
            'signature': signature,
        }

        if new_group not in self._permissions:
            self._groups.append(new_group)

    def add_permission(self, *, contract: str = IMPORT_WILDCARD, methods: Union[List[str], str] = IMPORT_WILDCARD):
        """
        Adds a valid contract and a valid methods to the permissions in the manifest.

        :param contract: a contract hash, group public key or '*'
        :type contract: str
        :param methods: a list of methods or '*'
        :type methods: Union[List[str], str]
        """

        if not isinstance(contract, str):
            return

        if not isinstance(methods, (str, list)):
            return

        from boa3.constants import IMPORT_WILDCARD
        # verifies if contract is a valid value
        if (not self._verify_is_valid_contract_hash(contract) and
                not self._verify_is_valid_public_key(contract) and
                contract != IMPORT_WILDCARD):
            return

        # verifies if methods is a valid value
        elif ((isinstance(methods, str) and methods != IMPORT_WILDCARD) or
              (isinstance(methods, list) and any(not isinstance(method, str) for method in methods))):
            return

        # if both values are the import wildcard, then it's not necessary to include it in the manifest
        if contract == '*' and methods == '*':
            return

        new_permission = {
            'contract': contract.lower(),
            'methods': methods,
        }

        if new_permission not in self._permissions:
            self._permissions.append(new_permission)

    @staticmethod
    def _verify_is_valid_public_key(public_key: str) -> bool:
        """
        Verifies if a given compressed public key is valid.

        :return: whether the given public key is valid or not
        """
        if public_key.startswith('03') or public_key.startswith('02'):
            try:
                from boa3 import constants
                if len(bytes.fromhex(public_key)) == constants.SIZE_OF_ECPOINT:
                    return True
            except ValueError:
                pass
        return False

    @staticmethod
    def _verify_is_valid_contract_hash(contract_hash: str) -> bool:
        """
        Verifies if a given contract hash is valid.

        :return: whether the given contract hash is valid or not
        """
        if contract_hash.startswith('0x'):
            try:
                from boa3.neo3.core.types import UInt160
                # if contract_hash is not a valid UInt160, it will raise a ValueError
                UInt160.from_string(contract_hash[2:])
                return True
            except ValueError:
                pass
        return False



from typing import Any, List, Union

from boa3.builtin.compile_time import NeoMetadata, metadata, public
from boa3.builtin.contract import Nep17TransferEvent
from boa3.builtin.interop import runtime, storage
from boa3.builtin.interop.contract import call_contract
from boa3.builtin.nativecontract.contractmanagement import ContractManagement
from boa3.builtin.nativecontract.gas import GAS as GAS_TOKEN
from boa3.builtin.nativecontract.neo import NEO as NEO_TOKEN
from boa3.builtin.type import UInt160


# -------------------------------------------
# METADATA
# -------------------------------------------


@metadata
def manifest_metadata() -> NeoMetadata:
    """
    Defines this smart contract's metadata information
    """
    meta = NeoMetadata()
    meta.supported_standards = ['NEP-17']
    meta.add_permission(methods=['onNEP17Payment'])

    meta.author = "Mirella Medeiros, Ricardo Prado and Lucas Uezu. COZ in partnership with Simpli"
    meta.description = "ICO Example"
    meta.email = "contact@coz.io"
    return meta


# -------------------------------------------
# Storage Key Prefixes
# -------------------------------------------


KYC_WHITELIST_PREFIX = b'KYCWhitelistApproved'
TOKEN_TOTAL_SUPPLY_PREFIX = b'TokenTotalSupply'
TRANSFER_ALLOWANCE_PREFIX = b'TransferAllowancePrefix_'


# -------------------------------------------
# TOKEN SETTINGS
# -------------------------------------------


# Script hash of the contract owner
TOKEN_OWNER = UInt160()

# Symbol of the Token
TOKEN_SYMBOL = 'ICO'

# Number of decimal places
TOKEN_DECIMALS = 8

# Initial Supply of tokens in the system
TOKEN_INITIAL_SUPPLY = 10_000_000 * 10 ** TOKEN_DECIMALS  # 10m total supply * 10^8 (decimals)

# -------------------------------------------
# Events
# -------------------------------------------


on_transfer = Nep17TransferEvent


# -------------------------------------------
# Methods
# -------------------------------------------


@public
def verify() -> bool:
    """
    When this contract address is included in the transaction signature,
    this method will be triggered as a VerificationTrigger to verify that the signature is correct.
    For example, this method needs to be called when withdrawing token from the contract.

    :return: whether the transaction signature is correct
    """
    return is_administrator()


def is_administrator() -> bool:
    """
    Validates if the invoker has administrative rights

    :return: whether the contract's invoker is an administrator
    """
    return runtime.check_witness(TOKEN_OWNER)


def is_valid_address(address: UInt160) -> bool:
    """
    Validates if the address passed through the kyc.

    :return: whether the given address is validated by kyc
    """
    return storage.get(KYC_WHITELIST_PREFIX + address).to_int() > 0


@public
def _deploy(data: Any, update: bool):
    """
    Initializes the storage when the smart contract is deployed.

    :return: whether the deploy was successful. This method must return True only during the smart contract's deploy.
    """
    if not update:
        storage.put(TOKEN_TOTAL_SUPPLY_PREFIX, TOKEN_INITIAL_SUPPLY)
        storage.put(TOKEN_OWNER, TOKEN_INITIAL_SUPPLY)

        on_transfer(None, TOKEN_OWNER, TOKEN_INITIAL_SUPPLY)


@public
def mint(amount: int) -> bool:
    """
    Mints new tokens

    :param amount: the amount of gas to be refunded
    :type amount: int
    :return: whether the refund was successful
    """
    assert amount >= 0
    if not is_administrator():
        return False

    if amount > 0:
        current_total_supply = total_supply()
        owner_balance = balance_of(TOKEN_OWNER)

        storage.put(TOKEN_TOTAL_SUPPLY_PREFIX, current_total_supply + amount)
        storage.put(TOKEN_OWNER, owner_balance + amount)

    on_transfer(None, TOKEN_OWNER, amount)
    post_transfer(None, TOKEN_OWNER, amount, None)
    return True


@public
def refund(address: UInt160, neo_amount: int, gas_amount: int) -> bool:
    """
    Refunds an address with given Neo and Gas

    :param address: the address that have the tokens
    :type address: UInt160
    :param neo_amount: the amount of neo to be refunded
    :type neo_amount: int
    :param gas_amount: the amount of gas to be refunded
    :type gas_amount: int
    :return: whether the refund was successful
    """
    assert len(address) == 20
    assert neo_amount > 0 or gas_amount > 0

    if not is_administrator():
        return False

    if neo_amount > 0:
        result = NEO_TOKEN.transfer(runtime.calling_script_hash, address, neo_amount)
        if not result:
            # due to a current limitation in the neo3-boa, changing the condition to `not result`
            # will result in a compiler error
            return False

    if gas_amount > 0:
        result = GAS_TOKEN.transfer(runtime.calling_script_hash, address, gas_amount)
        if not result:
            # due to a current limitation in the neo3-boa, changing the condition to `not result`
            # will result in a compiler error
            return False

    return True


# -------------------------------------------
# Public methods from NEP-17
# -------------------------------------------


@public(safe=True)
def symbol() -> str:
    """
    Gets the symbols of the token.

    This symbol should be short (3-8 characters is recommended), with no whitespace characters or new-lines and should
    be limited to the uppercase latin alphabet (i.e. the 26 letters used in English).
    This method must always return the same value every time it is invoked.

    :return: a short string symbol of the token managed in this contract.
    """
    return TOKEN_SYMBOL


@public(safe=True)
def decimals() -> int:
    """
    Gets the amount of decimals used by the token.

    E.g. 8, means to divide the token amount by 100,000,000 (10 ^ 8) to get its user representation.
    This method must always return the same value every time it is invoked.

    :return: the number of decimals used by the token.
    """
    return TOKEN_DECIMALS


@public(name='totalSupply', safe=True)
def total_supply() -> int:
    """
    Gets the total token supply deployed in the system.

    This number mustn't be in its user representation. E.g. if the total supply is 10,000,000 tokens, this method
    must return 10,000,000 * 10 ^ decimals.

    :return: the total token supply deployed in the system.
    """
    return storage.get(TOKEN_TOTAL_SUPPLY_PREFIX).to_int()


@public(name='balanceOf', safe=True)
def balance_of(account: UInt160) -> int:
    """
    Get the current balance of an address

    The parameter account should be a 20-byte address.

    :param account: the account address to retrieve the balance for
    :type account: UInt160

    :return: the token balance of the `account`
    :raise AssertionError: raised if `account` length is not 20.
    """
    assert len(account) == 20
    return storage.get(account).to_int()


@public
def transfer(from_address: UInt160, to_address: UInt160, amount: int, data: Any) -> bool:
    """
    Transfers a specified amount of NEP17 tokens from one account to another

    If the method succeeds, it must fire the `transfer` event and must return true, even if the amount is 0,
    or from and to are the same address.

    :param from_address: the address to transfer from
    :type from_address: UInt160
    :param to_address: the address to transfer to
    :type to_address: UInt160
    :param amount: the amount of NEP17 tokens to transfer
    :type amount: int
    :param data: whatever data is pertinent to the onPayment method
    :type data: Any

    :return: whether the transfer was successful
    :raise AssertionError: raised if `from_address` or `to_address` length is not 20 or if `amount` if less than zero.
    """
    # the parameters from and to should be 20-byte addresses. If not, this method should throw an exception.
    assert len(from_address) == 20 and len(to_address) == 20
    # the parameter amount must be greater than or equal to 0. If not, this method should throw an exception.
    assert amount >= 0

    # The function MUST return false if the from account balance does not have enough tokens to spend.
    from_balance = storage.get(from_address).to_int()
    if from_balance < amount:
        return False

    # The function should check whether the from address equals the caller contract hash.
    # If so, the transfer should be processed;
    # If not, the function should use the check_witness to verify the transfer.
    if from_address != runtime.calling_script_hash:
        if not runtime.check_witness(from_address):
            return False

    # skip balance changes if transferring to yourself or transferring 0 cryptocurrency
    if from_address != to_address and amount != 0:
        if from_balance == amount:
            storage.delete(from_address)
        else:
            storage.put(from_address, from_balance - amount)

        to_balance = storage.get(to_address).to_int()
        storage.put(to_address, to_balance + amount)

    # if the method succeeds, it must fire the transfer event
    on_transfer(from_address, to_address, amount)
    # if the to_address is a smart contract, it must call the contracts onPayment
    post_transfer(from_address, to_address, amount, data)
    # and then it must return true
    return True


def post_transfer(from_address: Union[UInt160, None], to_address: Union[UInt160, None], amount: int, data: Any):
    """
    Checks if the one receiving NEP17 tokens is a smart contract and if it's one the onPayment method will be called

    :param from_address: the address of the sender
    :type from_address: UInt160
    :param to_address: the address of the receiver
    :type to_address: UInt160
    :param amount: the amount of cryptocurrency that is being sent
    :type amount: int
    :param data: any pertinent data that might validate the transaction
    :type data: Any
    """
    if to_address is not None:
        contract = ContractManagement.get_contract(to_address)
        if contract is not None:
            call_contract(to_address, 'onNEP17Payment', [from_address, amount, data])


@public
def allowance(from_address: UInt160, to_address: UInt160) -> int:
    """
    Returns the amount of tokens that the to account can transfer from the from account.

    :param from_address: the address that have the tokens
    :type from_address: UInt160
    :param to_address: the address that is authorized to use the tokens
    :type to_address: UInt160

    :return: the amount of tokens that the `to` account can transfer from the `from` account
    :raise AssertionError: raised if `from_address` or `to_address` length is not 20.
    """
    # the parameters from and to should be 20-byte addresses. If not, this method should throw an exception.
    assert len(from_address) == 20 and len(to_address) == 20
    return storage.get(TRANSFER_ALLOWANCE_PREFIX + from_address + to_address).to_int()


@public(name='transferFrom')
def transfer_from(originator: UInt160, from_address: UInt160, to_address: UInt160, amount: int, data: Any) -> bool:
    """
    Transfers an amount from the `from` account to the `to` account if the `originator` has been approved to transfer
    the requested amount.

    :param originator: the address where the actual token is
    :type originator: UInt160
    :param from_address: the address to transfer from with originator's approval
    :type from_address: UInt160
    :param to_address: the address to transfer to
    :type to_address: UInt160
    :param amount: the amount of NEP17 tokens to transfer
    :type amount: int
    :param data: any pertinent data that might validate the transaction
    :type data: Any

    :return: whether the transfer was successful
    :raise AssertionError: raised if `from_address` or `to_address` length is not 20 or if `amount` if less than zero.
    """
    # the parameters from and to should be 20-byte addresses. If not, this method should throw an exception.
    assert len(originator) == 20 and len(from_address) == 20 and len(to_address) == 20
    # the parameter amount must be greater than or equal to 0. If not, this method should throw an exception.
    assert amount >= 0

    # The function should check whether the from address equals the caller contract hash.
    # If so, the transfer should be processed;
    # If not, the function should use the check_witness to verify the transfer.
    if from_address != runtime.calling_script_hash:
        if not runtime.check_witness(from_address):
            return False

    approved_transfer_amount = allowance(originator, from_address)
    if approved_transfer_amount < amount:
        return False

    originator_balance = balance_of(originator)
    if originator_balance < amount:
        return False

    # update allowance between originator and from
    if approved_transfer_amount == amount:
        storage.delete(TRANSFER_ALLOWANCE_PREFIX + originator + from_address)
    else:
        storage.put(TRANSFER_ALLOWANCE_PREFIX + originator + from_address, approved_transfer_amount - amount)

    # skip balance changes if transferring to yourself or transferring 0 cryptocurrency
    if amount != 0 and from_address != to_address:
        # update originator's balance
        if originator_balance == amount:
            storage.delete(originator)
        else:
            storage.put(originator, originator_balance - amount)

        # updates to's balance
        to_balance = storage.get(to_address).to_int()
        storage.put(to_address, to_balance + amount)

    # if the method succeeds, it must fire the transfer event
    on_transfer(from_address, to_address, amount)
    # if the to_address is a smart contract, it must call the contracts onPayment
    post_transfer(from_address, to_address, amount, data)
    # and then it must return true
    return True


@public
def approve(originator: UInt160, to_address: UInt160, amount: int) -> bool:
    """
    Approves the to account to transfer amount tokens from the originator account.

    :param originator: the address that have the tokens
    :type originator: UInt160
    :param to_address: the address that is authorized to use the tokens
    :type to_address: UInt160
    :param amount: the amount of NEP17 tokens to transfer
    :type amount: int

    :return: whether the approval was successful
    :raise AssertionError: raised if `originator` or `to_address` length is not 20 or if `amount` if less than zero.
    """
    assert len(originator) == 20 and len(to_address) == 20
    assert amount >= 0

    if not runtime.check_witness(originator):
        return False

    if originator == to_address:
        return False

    if not is_valid_address(originator) or not is_valid_address(to_address):
        # one of the address doesn't passed the kyc yet
        return False

    if balance_of(originator) < amount:
        return False

    storage.put(TRANSFER_ALLOWANCE_PREFIX + originator + to_address, amount)
    return True


# -------------------------------------------
# Public methods from KYC
# -------------------------------------------


@public
def kyc_register(addresses: List[UInt160]) -> int:
    """
    Includes the given addresses to the kyc whitelist

    :param addresses: a list with the addresses to be included
    :return: the number of included addresses
    """
    included_addresses = 0
    if is_administrator():
        for address in addresses:
            if len(address) == 20:
                kyc_key = KYC_WHITELIST_PREFIX + address
                storage.put(kyc_key, True)
                included_addresses += 1

    return included_addresses


@public
def kyc_remove(addresses: List[UInt160]) -> int:
    """
    Removes the given addresses from the kyc whitelist

    :param addresses: a list with the addresses to be removed
    :return: the number of removed addresses
    """
    removed_addresses = 0
    if is_administrator():
        for address in addresses:
            if len(address) == 20:
                kyc_key = KYC_WHITELIST_PREFIX + address
                storage.delete(kyc_key)
                removed_addresses += 1

    return removed_addresses



import hashlib

from boa3 import constants
from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.model.type.type import Type
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo3.contracts.namedcurve import NamedCurve
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.testengine import TestEngine


class TestCryptoLibClass(BoaTest):
    default_folder: str = 'test_sc/native_test/cryptolib'
    ecpoint_init = (
        Opcode.CONVERT + Type.bytes.stack_item
        + Opcode.DUP
        + Opcode.ISNULL
        + Opcode.JMPIF + Integer(8).to_byte_array()
        + Opcode.DUP
        + Opcode.SIZE
        + Opcode.PUSHINT8 + Integer(33).to_byte_array(signed=True)
        + Opcode.JMPEQ + Integer(3).to_byte_array()
        + Opcode.THROW
    )

    def test_get_hash(self):
        path = self.get_contract_path('GetHash.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(constants.CRYPTO_SCRIPT, result)

    def test_ripemd160_str(self):
        path = self.get_contract_path('Ripemd160Str.py')
        engine = TestEngine()
        expected_result = hashlib.new('ripemd160', b'unit test')
        result = self.run_smart_contract(engine, path, 'Main', 'unit test')
        self.assertEqual(expected_result.digest(), result)

        expected_result = hashlib.new('ripemd160', b'')
        result = self.run_smart_contract(engine, path, 'Main', '')
        self.assertEqual(expected_result.digest(), result)

    def test_ripemd160_int(self):
        path = self.get_contract_path('Ripemd160Int.py')
        engine = TestEngine()
        expected_result = hashlib.new('ripemd160', Integer(10).to_byte_array())
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(expected_result.digest(), result)

    def test_ripemd160_bool(self):
        path = self.get_contract_path('Ripemd160Bool.py')
        engine = TestEngine()
        expected_result = hashlib.new('ripemd160', Integer(1).to_byte_array())
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(expected_result.digest(), result)

    def test_ripemd160_bytes(self):
        path = self.get_contract_path('Ripemd160Bytes.py')
        engine = TestEngine()
        expected_result = hashlib.new('ripemd160', b'unit test')
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(expected_result.digest(), result)

    def test_ripemd160_too_many_parameters(self):
        path = self.get_contract_path('Ripemd160TooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_ripemd160_too_few_parameters(self):
        path = self.get_contract_path('Ripemd160TooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_sha256_str(self):
        path = self.get_contract_path('Sha256Str.py')
        engine = TestEngine()
        expected_result = hashlib.sha256(b'unit test')
        result = self.run_smart_contract(engine, path, 'Main', 'unit test')
        self.assertEqual(expected_result.digest(), result)

        expected_result = hashlib.sha256(b'')
        result = self.run_smart_contract(engine, path, 'Main', '')
        self.assertEqual(expected_result.digest(), result)

    def test_sha256_int(self):
        path = self.get_contract_path('Sha256Int.py')
        engine = TestEngine()
        expected_result = hashlib.sha256(Integer(10).to_byte_array())
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(expected_result.digest(), result)

    def test_sha256_bool(self):
        path = self.get_contract_path('Sha256Bool.py')
        engine = TestEngine()
        expected_result = hashlib.sha256(Integer(1).to_byte_array())
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(expected_result.digest(), result)

    def test_sha256_bytes(self):
        path = self.get_contract_path('Sha256Bytes.py')
        engine = TestEngine()
        expected_result = hashlib.sha256(b'unit test')
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(expected_result.digest(), result)

    def test_sha256_too_many_parameters(self):
        path = self.get_contract_path('Sha256TooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_sha256_too_few_parameters(self):
        path = self.get_contract_path('Sha256TooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_verify_with_ecdsa(self):
        path = self.get_contract_path('VerifyWithECDsa.py')
        Boa3.compile(path)

    def test_verify_with_ecdsa_secp256r1_str(self):
        byte_input1 = b'0123456789ABCDEFGHIJKLMNOPQRSTUVW'
        byte_input2 = b'signature'
        string = b'unit test'
        named_curve = Integer(NamedCurve.SECP256R1).to_byte_array(signed=True, min_length=1)

        expected_output = (
            Opcode.PUSHINT8 + named_curve
            + Opcode.PUSHDATA1
            + Integer(len(byte_input2)).to_byte_array(min_length=1)
            + byte_input2
            + Opcode.PUSHDATA1
            + Integer(len(byte_input1)).to_byte_array(min_length=1)
            + byte_input1
            + self.ecpoint_init
            + Opcode.PUSHDATA1
            + Integer(len(string)).to_byte_array(min_length=1)
            + string
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.DROP
            + Opcode.RET
        )

        path = self.get_contract_path('VerifyWithECDsaSecp256r1Str.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_verify_with_ecdsa_secp256r1_bool(self):
        byte_input1 = b'0123456789ABCDEFGHIJKLMNOPQRSTUVW'
        byte_input2 = b'signature'
        named_curve = Integer(NamedCurve.SECP256R1).to_byte_array(signed=True, min_length=1)

        expected_output = (
            Opcode.PUSHINT8 + named_curve
            + Opcode.PUSHDATA1
            + Integer(len(byte_input2)).to_byte_array(min_length=1)
            + byte_input2
            + Opcode.PUSHDATA1
            + Integer(len(byte_input1)).to_byte_array(min_length=1)
            + byte_input1
            + self.ecpoint_init
            + Opcode.PUSHF
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.DROP
            + Opcode.RET
        )

        path = self.get_contract_path('VerifyWithECDsaSecp256r1Bool.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_verify_with_ecdsa_secp256r1_int(self):
        byte_input1 = b'0123456789ABCDEFGHIJKLMNOPQRSTUVW'
        byte_input2 = b'signature'
        named_curve = Integer(NamedCurve.SECP256R1).to_byte_array(signed=True, min_length=1)

        expected_output = (
            Opcode.PUSHINT8 + named_curve
            + Opcode.PUSHDATA1
            + Integer(len(byte_input2)).to_byte_array(min_length=1)
            + byte_input2
            + Opcode.PUSHDATA1
            + Integer(len(byte_input1)).to_byte_array(min_length=1)
            + byte_input1
            + self.ecpoint_init
            + Opcode.PUSH10
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.DROP
            + Opcode.RET
        )

        path = self.get_contract_path('VerifyWithECDsaSecp256r1Int.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_verify_with_ecdsa_secp256r1_bytes(self):
        byte_input1 = b'0123456789ABCDEFGHIJKLMNOPQRSTUVW'
        byte_input2 = b'signature'
        string = b'unit test'
        named_curve = Integer(NamedCurve.SECP256R1).to_byte_array(signed=True, min_length=1)

        expected_output = (
            Opcode.PUSHINT8 + named_curve
            + Opcode.PUSHDATA1
            + Integer(len(byte_input2)).to_byte_array(min_length=1)
            + byte_input2
            + Opcode.PUSHDATA1
            + Integer(len(byte_input1)).to_byte_array(min_length=1)
            + byte_input1
            + self.ecpoint_init
            + Opcode.PUSHDATA1
            + Integer(len(string)).to_byte_array(min_length=1)
            + string
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.DROP
            + Opcode.RET
        )

        path = self.get_contract_path('VerifyWithECDsaSecp256r1Bytes.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_verify_with_ecdsa_secp256r1_mismatched_type(self):
        path = self.get_contract_path('VerifyWithECDsaSecp256r1MismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_verify_with_ecdsa_secp256k1_str(self):
        byte_input1 = b'0123456789ABCDEFGHIJKLMNOPQRSTUVW'
        byte_input2 = b'signature'
        string = b'unit test'
        named_curve = Integer(NamedCurve.SECP256K1).to_byte_array(signed=True, min_length=1)

        expected_output = (
            Opcode.PUSHINT8 + named_curve
            + Opcode.PUSHDATA1
            + Integer(len(byte_input2)).to_byte_array(min_length=1)
            + byte_input2
            + Opcode.PUSHDATA1
            + Integer(len(byte_input1)).to_byte_array(min_length=1)
            + byte_input1
            + self.ecpoint_init
            + Opcode.PUSHDATA1
            + Integer(len(string)).to_byte_array(min_length=1)
            + string
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.DROP
            + Opcode.RET
        )

        path = self.get_contract_path('VerifyWithECDsaSecp256k1Str.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_verify_with_ecdsa_secp256k1_bool(self):
        byte_input1 = b'0123456789ABCDEFGHIJKLMNOPQRSTUVW'
        byte_input2 = b'signature'
        named_curve = Integer(NamedCurve.SECP256K1).to_byte_array(signed=True, min_length=1)

        expected_output = (
            Opcode.PUSHINT8 + named_curve
            + Opcode.PUSHDATA1
            + Integer(len(byte_input2)).to_byte_array(min_length=1)
            + byte_input2
            + Opcode.PUSHDATA1
            + Integer(len(byte_input1)).to_byte_array(min_length=1)
            + byte_input1
            + self.ecpoint_init
            + Opcode.PUSHF
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.DROP
            + Opcode.RET
        )

        path = self.get_contract_path('VerifyWithECDsaSecp256k1Bool.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_verify_with_ecdsa_secp256k1_int(self):
        byte_input1 = b'0123456789ABCDEFGHIJKLMNOPQRSTUVW'
        byte_input2 = b'signature'
        named_curve = Integer(NamedCurve.SECP256K1).to_byte_array(signed=True, min_length=1)

        expected_output = (
            Opcode.PUSHINT8 + named_curve
            + Opcode.PUSHDATA1
            + Integer(len(byte_input2)).to_byte_array(min_length=1)
            + byte_input2
            + Opcode.PUSHDATA1
            + Integer(len(byte_input1)).to_byte_array(min_length=1)
            + byte_input1
            + self.ecpoint_init
            + Opcode.PUSH10
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.DROP
            + Opcode.RET
        )

        path = self.get_contract_path('VerifyWithECDsaSecp256k1Int.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_verify_with_ecdsa_secp256k1_bytes(self):
        byte_input1 = b'0123456789ABCDEFGHIJKLMNOPQRSTUVW'
        byte_input2 = b'signature'
        string = b'unit test'
        named_curve = Integer(NamedCurve.SECP256K1).to_byte_array(signed=True, min_length=1)

        expected_output = (
            Opcode.PUSHINT8 + named_curve
            + Opcode.PUSHDATA1
            + Integer(len(byte_input2)).to_byte_array(min_length=1)
            + byte_input2
            + Opcode.PUSHDATA1
            + Integer(len(byte_input1)).to_byte_array(min_length=1)
            + byte_input1
            + self.ecpoint_init
            + Opcode.PUSHDATA1
            + Integer(len(string)).to_byte_array(min_length=1)
            + string
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.DROP
            + Opcode.RET
        )

        path = self.get_contract_path('VerifyWithECDsaSecp256k1Bytes.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_verify_with_ecdsa_secp256k1_mismatched_type(self):
        path = self.get_contract_path('VerifyWithECDsaSecp256k1MismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_murmur32(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00'
            + b'\x02'
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.RET
        )

        path = self.get_contract_path('Murmur32.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)



from typing import Dict

from boa3.model.builtin.interop.nativecontract import ContractManagementMethod
from boa3.model.variable import Variable


class HasMethod(ContractManagementMethod):

    def __init__(self):
        from boa3.model.type.collection.sequence.uint160type import UInt160Type
        from boa3.model.type.type import Type

        identifier = 'has_method'
        syscall = 'hasMethod'
        args: Dict[str, Variable] = {
            'hash': Variable(UInt160Type.build()),
            'method': Variable(Type.str),
            'parameter_count': Variable(Type.int)
        }

        super().__init__(identifier, syscall, args, return_type=Type.bool)



import ast
from typing import Dict

from boa3.model.builtin.interop.nativecontract import Nep17Method
from boa3.model.variable import Variable


class TransferMethod(Nep17Method):

    def __init__(self, contract_script_hash: bytes):
        from boa3.model.type.type import Type
        from boa3.model.type.collection.sequence.uint160type import UInt160Type

        identifier = 'transfer'
        native_identifier = 'transfer'
        args: Dict[str, Variable] = {
            'from_address': Variable(UInt160Type.build()),
            'to_address': Variable(UInt160Type.build()),
            'amount': Variable(Type.int),
            'data': Variable(Type.any),
        }

        data_default = ast.parse("{0}".format(Type.any.default_value)
                                 ).body[0].value

        super().__init__(identifier, native_identifier, args, defaults=[data_default],
                         return_type=Type.bool, script_hash=contract_script_hash)



import json

from boa3 import constants
from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.neo.vm.type.String import String
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.contract.neomanifeststruct import NeoManifestStruct
from boa3_test.tests.test_classes.testengine import TestEngine


class TestContractManagementContract(BoaTest):
    default_folder: str = 'test_sc/native_test/contractmanagement'

    def test_get_hash(self):
        path = self.get_contract_path('GetHash.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(constants.MANAGEMENT_SCRIPT, result)

    def test_get_minimum_deployment_fee(self):
        path = self.get_contract_path('GetMinimumDeploymentFee.py')
        engine = TestEngine()

        minimum_cost = 10 * 10 ** 8  # minimum deployment cost is 10 GAS right now
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(minimum_cost, result)

    def test_get_minimum_deployment_fee_too_many_parameters(self):
        path = self.get_contract_path('GetMinimumDeploymentFeeTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_get_contract(self):
        path = self.get_contract_path('GetContract.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main', bytes(20))
        self.assertIsNone(result)

        call_contract_path = self.get_contract_path('test_sc/arithmetic_test', 'Addition.py')
        nef, manifest = self.get_bytes_output(call_contract_path)
        self.run_smart_contract(engine, call_contract_path, 'add', 1, 2)
        call_hash = engine.executed_script_hash.to_array()
        call_contract_path = call_contract_path.replace('.py', '.nef')

        engine = TestEngine()
        engine.add_contract(call_contract_path)

        result = self.run_smart_contract(engine, path, 'main', call_hash)
        self.assertEqual(5, len(result))
        self.assertEqual(call_hash, result[2])
        self.assertEqual(nef, result[3])
        manifest_struct = NeoManifestStruct.from_json(manifest)
        self.assertEqual(manifest_struct, result[4])

    def test_has_method(self):
        path = self.get_contract_path('HasMethod.py')
        engine = TestEngine()

        test_method = 'add'
        test_parameter_count = 2
        result = self.run_smart_contract(engine, path, 'main', bytes(20), test_method, test_parameter_count)
        self.assertEqual(False, result)

        call_contract_path = self.get_contract_path('test_sc/arithmetic_test', 'Addition.py')
        self.run_smart_contract(engine, call_contract_path, 'add', 1, 2)
        call_hash = engine.executed_script_hash.to_array()

        result = self.run_smart_contract(engine, path, 'main', call_hash, test_method, test_parameter_count)
        self.assertEqual(True, result)

    def test_deploy_contract(self):
        path = self.get_contract_path('DeployContract.py')
        call_contract_path = self.get_contract_path('test_sc/arithmetic_test', 'Addition.py')
        Boa3.compile_and_save(call_contract_path)
        self.compile_and_save(path)

        nef_file, manifest = self.get_bytes_output(call_contract_path)
        arg_manifest = String(json.dumps(manifest, separators=(',', ':'))).to_bytes()

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', nef_file, arg_manifest, None)

        self.assertEqual(5, len(result))
        self.assertEqual(nef_file, result[3])
        manifest_struct = NeoManifestStruct.from_json(manifest)
        self.assertEqual(manifest_struct, result[4])

    def test_deploy_contract_data_deploy(self):
        path = self.get_contract_path('DeployContract.py')
        call_contract_path = self.get_contract_path('boa3_test/test_sc/interop_test/contract', 'NewContract.py')
        self.compile_and_save(call_contract_path)

        nef_file, manifest = self.get_bytes_output(call_contract_path)
        arg_manifest = String(json.dumps(manifest, separators=(',', ':'))).to_bytes()

        engine = TestEngine()
        data = 'some sort of data'
        result = self.run_smart_contract(engine, path, 'Main', nef_file, arg_manifest, data)

        self.assertEqual(5, len(result))
        self.assertEqual(nef_file, result[3])
        manifest_struct = NeoManifestStruct.from_json(manifest)
        self.assertEqual(manifest_struct, result[4])

        notifies = engine.get_events('notify')
        self.assertEqual(2, len(notifies))
        self.assertEqual(False, notifies[0].arguments[0])  # not updated
        self.assertEqual(data, notifies[1].arguments[0])   # data
        result = self.run_smart_contract(engine, call_contract_path, 'main')
        self.assertEqual(data, result)

    def test_deploy_contract_too_many_parameters(self):
        path = self.get_contract_path('DeployContractTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_deploy_contract_too_few_parameters(self):
        path = self.get_contract_path('DeployContractTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_update_contract(self):
        path = self.get_contract_path('UpdateContract.py')
        engine = TestEngine()
        with self.assertRaisesRegex(TestExecutionException, f'{self.CANT_FIND_METHOD_MSG_PREFIX} : new_method'):
            self.run_smart_contract(engine, path, 'new_method')

        new_path = self.get_contract_path('test_sc/interop_test', 'UpdateContract.py')
        self.compile_and_save(new_path)
        new_nef, new_manifest = self.get_bytes_output(new_path)
        arg_manifest = String(json.dumps(new_manifest, separators=(',', ':'))).to_bytes()

        result = self.run_smart_contract(engine, path, 'update', new_nef, arg_manifest, None)
        self.assertIsVoid(result)

        result = self.run_smart_contract(engine, path, 'new_method')
        self.assertEqual(42, result)

    def test_update_contract_data_deploy(self):
        path = self.get_contract_path('UpdateContract.py')
        engine = TestEngine()
        with self.assertRaisesRegex(TestExecutionException, f'{self.CANT_FIND_METHOD_MSG_PREFIX} : new_method'):
            self.run_smart_contract(engine, path, 'new_method')

        new_path = self.get_contract_path('test_sc/interop_test', 'UpdateContract.py')
        self.compile_and_save(new_path)
        new_nef, new_manifest = self.get_bytes_output(new_path)
        arg_manifest = String(json.dumps(new_manifest, separators=(',', ':'))).to_bytes()

        data = 'this function was deployed'
        result = self.run_smart_contract(engine, path, 'update', new_nef, arg_manifest, data)
        self.assertIsVoid(result)
        notifies = engine.get_events('notify')
        self.assertEqual(2, len(notifies))
        self.assertEqual(True, notifies[0].arguments[0])
        self.assertEqual(data, notifies[1].arguments[0])

    def test_update_contract_too_many_parameters(self):
        path = self.get_contract_path('UpdateContractTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_update_contract_too_few_parameters(self):
        path = self.get_contract_path('UpdateContractTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_destroy_contract(self):
        path = self.get_contract_path('DestroyContract.py')
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)

        script_hash = engine.executed_script_hash.to_array()
        call_contract_path = self.get_contract_path('boa3_test/test_sc/interop_test/contract', 'CallScriptHash.py')
        with self.assertRaisesRegex(TestExecutionException, f'^{self.CALLED_CONTRACT_DOES_NOT_EXIST_MSG}'):
            self.run_smart_contract(engine, call_contract_path, 'Main',
                                    script_hash, 'Main', [])

    def test_destroy_contract_too_many_parameters(self):
        path = self.get_contract_path('DestroyContractTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)



from boa3.builtin.compile_time import public
from boa3.builtin.type import UInt160


@public
def main() -> UInt160:
    bytes_value = b'0123456789abcdefghijklmnopqrstuvwxyz'
    len20 = bytes_value[:20]
    return UInt160(len20)



import module1
import module2

from boa3.builtin.compile_time import public


@public
def main() -> int:
    return 1 + module2.sample_method()


def sample_method() -> int:
    return module1.sample_method()



from __future__ import annotations

from typing import Any, Dict

from boa3_test.tests.test_classes.contract.neoabistruct import NeoAbiStruct
from boa3_test.tests.test_classes.contract.neopermissionsstruct import NeoPermissionsStruct
from boa3_test.tests.test_classes.contract.neostruct import NeoStruct


class NeoManifestStruct(NeoStruct):
    _name_field = 'name'
    _groups_field = 'groups'
    _supported_standards_field = 'supportedstandards'
    _abi_field = 'abi'
    _permissions_field = 'permissions'
    _trusts_field = 'trusts'
    _features_field = 'features'
    _extra_field = 'extra'

    @classmethod
    def from_json(cls, json: Dict[str, Any]) -> NeoManifestStruct:
        required_fields = [cls._name_field,
                           cls._groups_field,
                           cls._supported_standards_field,
                           cls._abi_field,
                           cls._permissions_field,
                           cls._trusts_field,
                           cls._features_field,
                           cls._extra_field
                           ]
        cls._validate_json(json, required_fields)

        struct = cls()
        struct.append(json[cls._name_field])
        struct.append(json[cls._groups_field])
        struct.append({})  # TODO: features aren't implemented yet
        struct.append(json[cls._supported_standards_field])
        struct.append(NeoAbiStruct.from_json(json[cls._abi_field]))
        struct.append([NeoPermissionsStruct.from_json(permission) for permission in json[cls._permissions_field]])
        struct.append(json[cls._trusts_field])
        extras = json[cls._extra_field]
        struct.append(extras if extras is not None else "null")

        return struct



import ast
from typing import Dict, List, Tuple

from boa3.model import set_internal_call
from boa3.model.builtin.interop.interopmethod import InteropMethod
from boa3.model.builtin.interop.runtime.notificationtype import NotificationType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode import OpcodeHelper
from boa3.neo.vm.opcode.Opcode import Opcode


class GetNotificationsMethod(InteropMethod):

    def __init__(self, notification_type: NotificationType):
        from boa3.model.type.collection.sequence.uint160type import UInt160Type
        from boa3.model.type.type import Type

        identifier = 'get_notifications'
        syscall = 'System.Runtime.GetNotifications'
        uint160 = UInt160Type.build()

        args: Dict[str, Variable] = {'script_hash': Variable(Type.optional.build(uint160))}
        args_default = set_internal_call(ast.parse("{0}".format(Type.none.default_value)
                                                   ).body[0].value)

        super().__init__(identifier, syscall, args, [args_default],
                         return_type=Type.list.build([notification_type]))

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        from boa3.neo.vm.type.Integer import Integer
        from boa3.model.type.type import Type
        jmp_place_holder = (Opcode.JMP, b'\x01')

        verify_arg_is_null = [
            (Opcode.DUP, b''),
            (Opcode.JMPIFNOT, jmp_place_holder),
        ]

        arg_is_not_null = [
            (Opcode.DUP, b''),
            (Opcode.CONVERT, Type.int.stack_item),
            (Opcode.PUSH0, b''),
            (Opcode.NUMEQUAL, b''),
            (Opcode.JMPIFNOT, Integer(4).to_byte_array(signed=True)),
            (Opcode.DROP, b''),
            (Opcode.PUSHNULL, b'')
        ]

        from boa3.compiler.codegenerator import get_bytes_count
        jmp_to_convert = OpcodeHelper.get_jump_and_data(Opcode.JMPIFNOT, get_bytes_count(arg_is_not_null), True)
        verify_arg_is_null[-1] = jmp_to_convert

        return (
            verify_arg_is_null +
            arg_is_not_null +
            super()._opcode
        )



from boa3.builtin.type import UInt160, UInt256


class Block:
    """
    Represents a block.

    :ivar hash: a unique identifier based on the unsigned data portion of the object
    :vartype hash: UInt256
    :ivar version: the data structure version of the block
    :vartype version: int
    :ivar previous_hash: the hash of the previous block
    :vartype previous_hash: UInt256
    :ivar merkle_root: the merkle root of the transactions
    :vartype merkle_root: UInt256
    :ivar timestamp: UTC timestamp of the block in milliseconds
    :vartype timestamp: int
    :ivar nonce: a random number used once in the cryptography
    :vartype nonce: int
    :ivar index: the index of the block
    :vartype index: int
    :ivar primary_index: the primary index of the consensus node that generated this block
    :vartype primary_index: int
    :ivar next_consensus: the script hash of the consensus nodes that generates the next block
    :vartype next_consensus: UInt160
    :ivar transaction_count: the number of transactions on this block
    :vartype transaction_count: int
    """

    def __init__(self):
        self.hash: UInt256 = UInt256()
        self.version: int = 0
        self.previous_hash: UInt256 = UInt256()
        self.merkle_root: UInt256 = UInt256()
        self.timestamp: int = 0
        self.nonce: int = 0
        self.index: int = 0
        self.primary_index: int = 0
        self.next_consensus: UInt160 = UInt160()
        self.transaction_count: int = 0



from typing import Dict

from boa3.model.builtin.method.builtinevent import IBuiltinEvent
from boa3.model.variable import Variable


class Nep5TransferEvent(IBuiltinEvent):

    def __init__(self):
        from boa3.model.type.type import Type
        identifier = 'Nep5TransferEvent'
        args: Dict[str, Variable] = {
            'from_addr': Variable(Type.bytes),
            'to_addr': Variable(Type.bytes),
            'amount': Variable(Type.int)
        }
        super().__init__(identifier, args)
        self.name = 'transfer'



from boa3.builtin.compile_time import public


@public
def omitted_values() -> bytes:
    a = b'unit_test'
    return a[::-2]


@public
def omitted_start() -> bytes:
    a = b'unit_test'
    return a[:5:-2]


@public
def omitted_end() -> bytes:
    a = b'unit_test'
    return a[2::-2]


@public
def negative_start() -> bytes:
    a = b'unit_test'
    return a[-6::-2]


@public
def negative_end() -> bytes:
    a = b'unit_test'
    return a[:-1:-2]


@public
def negative_really_low_start() -> bytes:
    a = b'unit_test'
    return a[-999::-2]


@public
def negative_really_low_end() -> bytes:
    a = b'unit_test'
    return a[:-999:-2]


@public
def really_high_start() -> bytes:
    a = b'unit_test'
    return a[999::-2]


@public
def really_high_end() -> bytes:
    a = b'unit_test'
    return a[:999:-2]



import ast
from abc import ABC
from typing import Dict, List, Optional, Tuple

from boa3.model.callable import Callable
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.type.itype import IType
from boa3.model.variable import Variable
from boa3.neo.vm.opcode.Opcode import Opcode


class IBuiltinCallable(Callable, IdentifiedSymbol, ABC):
    def __init__(self, identifier: str, args: Dict[str, Variable] = None,
                 vararg: Optional[Tuple[str, Variable]] = None,
                 kwargs: Optional[Dict[str, Variable]] = None,
                 defaults: List[ast.AST] = None, return_type: IType = None):
        super().__init__(args, vararg, kwargs, defaults, return_type)
        self._identifier = identifier
        self._generated_opcode = None
        self.defined_by_entry = False  # every builtin symbol must have this variable set as False

    @property
    def opcode(self) -> List[Tuple[Opcode, bytes]]:
        """
        Gets the opcode for the method.

        :return: the opcode and its data if exists. None otherwise.
        """
        # don't need to recalculate for every time this property is called
        if self._generated_opcode is None:
            self._generated_opcode = self._opcode
        return self._generated_opcode

    def reset(self):
        # reset the opcodes to ensure the correct output when calling consecutive compilations
        self._generated_opcode = None
        self.reset_calls()

    @property
    def _opcode(self) -> List[Tuple[Opcode, bytes]]:
        return []

    @property
    def identifier(self) -> str:
        return self._identifier



from typing import List

from boa3.model.builtin.builtin import Builtin
from boa3.model.builtin.interop.interop import Interop
from boa3.model.builtin.native import *
from boa3.model.builtin.native.oracleclass import OracleClass
from boa3.model.identifiedsymbol import IdentifiedSymbol
from boa3.model.imports.package import Package


class NativeContract:
    # Class Interfaces
    ContractManagement = ContractManagementClass.build()
    CryptoLib = CryptoLibClass.build()
    GAS = GasClass.build()
    Ledger = LedgerClass.build()
    NEO = NeoClass.build()
    Oracle = OracleClass.build()
    Policy = PolicyClass.build()
    RoleManagement = RoleManagementClass.build()
    StdLib = StdLibClass.build()

    # region Packages

    ContractManagementModule = Package(identifier=ContractManagement.identifier.lower(),
                                       types=[ContractManagement,
                                              Interop.ContractType,
                                              Builtin.UInt160])

    CryptoLibModule = Package(identifier=CryptoLib.identifier.lower(),
                              types=[CryptoLib,
                                     Interop.NamedCurveType])

    GasModule = Package(identifier=GAS.identifier.lower(),
                        types=[GAS]
                        )

    LedgerModule = Package(identifier=Ledger.identifier.lower(),
                           types=[Ledger,
                                  Interop.BlockType,
                                  Interop.TransactionType]
                           )

    NeoModule = Package(identifier=NEO.identifier.lower(),
                        types=[NEO]
                        )

    OracleModule = Package(identifier=Oracle.identifier.lower(),
                           types=[Oracle]
                           )

    PolicyModule = Package(identifier=Policy.identifier.lower(),
                           types=[Policy]
                           )

    RoleManagementModule = Package(identifier=RoleManagement.identifier.lower(),
                                   types=[RoleManagement,
                                          Interop.RoleType]
                                   )

    StdLibModule = Package(identifier=StdLib.identifier.lower(),
                           types=[StdLib]
                           )

    # endregion

    package_symbols: List[IdentifiedSymbol] = [
        ContractManagementModule,
        CryptoLibModule,
        GasModule,
        LedgerModule,
        NeoModule,
        OracleModule,
        PolicyModule,
        RoleManagementModule,
        StdLibModule
    ]



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Dict, List


class NeoStruct(list, ABC):

    @classmethod
    @abstractmethod
    def from_json(cls, json: Dict[str, Any]) -> NeoStruct:
        pass

    @classmethod
    def _validate_json(cls, json: Dict[str, Any], required_fields: List[str]):
        required = set(required_fields)
        if not required.issubset(json.keys()):
            raise ValueError

    @classmethod
    def _get_param_info(cls, json: Dict[str, Any]) -> list:
        _name = 'name'
        _type = 'type'

        required_fields = [_name, _type]
        cls._validate_json(json, required_fields)

        name = json[_name]
        from boa3.neo.vm.type.ContractParameterType import ContractParameterType
        try:
            param_type = ContractParameterType._get_by_name(json[_type])
        except BaseException:
            param_type = None

        return [name, param_type]



from boa3.builtin.compile_time import public
from boa3.builtin.interop.runtime import notify


@public
def main() -> bytes:

    m = b'\x01\x02\x03\x04\x05\x06\x07\x08'

    s2 = m[:4]

    j = 2
    k = 4

    s3 = m[j:k]

    notify(s3)

    s4 = m[get_slice_start():get_slice_end()]

    notify(s4)

    ind = [1, 3, 4, 5]

    s6 = m[get_slice_start():ind[2]]

    notify(s6)

    res = s6 + s4 + s2 + s3

    notify(res)

    return res


def get_slice_start() -> int:

    return 1


def get_slice_end() -> int:
    return 6



from boa3.builtin.compile_time import public
from boa3.builtin.type import UInt256


@public
def main() -> UInt256:
    bytes_value = b'0123456789abcdefghijklmnopqrstuvwxyz'
    len32 = bytes_value[:32]
    return UInt256(len32)



from boa3_test.tests.test_classes.witnessscope import WitnessScope
from test_runner.neoxp.command.neoexpresscommand import NeoExpressCommand


def deploy(nef_path: str, account: str, witness_scope: WitnessScope = None,
           force: bool = False, trace: bool = False) -> NeoExpressCommand:
    options = {}
    if isinstance(witness_scope, WitnessScope):
        options['--witness-scope'] = witness_scope.name
    if trace:
        options['--trace'] = ''
    if force:
        options['--force'] = ''

    return NeoExpressCommand('contract deploy', [nef_path, account], options)


def run(contract_id: str, method: str, *args: str, account: str,
        witness_scope: WitnessScope = None, trace: bool = False) -> NeoExpressCommand:
    options = {}
    if isinstance(witness_scope, WitnessScope):
        options['--witness-scope'] = witness_scope.name
    if isinstance(account, str):
        options['--account'] = account
    if trace:
        options['--trace'] = ''

    arguments = [contract_id, method]
    arguments.extend(args)

    return NeoExpressCommand('contract run', arguments, options)



from typing import Dict

from boa3.model.builtin.interop.nativecontract import StdLibMethod
from boa3.model.variable import Variable


class MemoryCompareMethod(StdLibMethod):

    def __init__(self):
        from boa3.model.type.type import Type
        from boa3.model.type.primitive.bytestringtype import ByteStringType

        identifier = 'memory_compare'
        syscall = 'memoryCompare'
        byte_string_type = ByteStringType.build()

        args: Dict[str, Variable] = {
            'mem1': Variable(byte_string_type),
            'mem2': Variable(byte_string_type)
        }

        super().__init__(identifier, syscall, args, return_type=Type.int)



from boa3.boa3 import Boa3
from boa3.exception import CompilerError
from boa3.model.builtin.interop.interop import Interop
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3.neo3.contracts import CallFlags
from boa3_test.tests.boa_test import BoaTest


class TestRoleInterop(BoaTest):
    default_folder: str = 'test_sc/interop_test/role'

    def test_get_designated_by_role(self):
        call_flags = Integer(CallFlags.ALL).to_byte_array(signed=True, min_length=1)
        method = String(Interop.GetDesignatedByRole.method_name).to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x02'
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.RET
        )

        path = self.get_contract_path('GetDesignatedByRole.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_get_designated_by_role_too_many_parameters(self):
        path = self.get_contract_path('GetDesignatedByRoleTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_get_designated_by_role_too_few_parameters(self):
        path = self.get_contract_path('GetDesignatedByRoleTooFewArguments.py')
        self.assertCompilerLogs(CompilerError.UnfilledArgument, path)

    def test_import_role(self):
        call_flags = Integer(CallFlags.ALL).to_byte_array(signed=True, min_length=1)
        method = String(Interop.GetDesignatedByRole.method_name).to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x02'
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.RET
        )

        path = self.get_contract_path('ImportRole.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_import_interop_role(self):
        call_flags = Integer(CallFlags.ALL).to_byte_array(signed=True, min_length=1)
        method = String(Interop.GetDesignatedByRole.method_name).to_bytes()

        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x02'
            + Opcode.LDARG1
            + Opcode.LDARG0
            + Opcode.CALLT + b'\x00\x00'
            + Opcode.RET
        )

        path = self.get_contract_path('ImportInteropRole.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)



from boa3 import constants
from boa3.boa3 import Boa3
from boa3.exception import CompilerError, CompilerWarning
from boa3.model.builtin.interop.interop import Interop
from boa3.neo import to_script_hash
from boa3.neo.vm.opcode.Opcode import Opcode
from boa3.neo.vm.type.Integer import Integer
from boa3.neo.vm.type.String import String
from boa3.neo3.contracts import TriggerType
from boa3_test.tests.boa_test import BoaTest
from boa3_test.tests.test_classes.TestExecutionException import TestExecutionException
from boa3_test.tests.test_classes.testengine import TestEngine


class TestRuntimeInterop(BoaTest):
    default_folder: str = 'test_sc/interop_test/runtime'

    def test_check_witness(self):
        path = self.get_contract_path('CheckWitness.py')
        account = to_script_hash(b'NiNmXL8FjEUEs1nfX9uHFBNaenxDHJtmuB')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', account)
        self.assertEqual(False, result)

        engine.add_signer_account(account)
        result = self.run_smart_contract(engine, path, 'Main', account)
        self.assertEqual(True, result)

    def test_contract_with_check_witness(self):
        path = self.get_contract_path('test_sc/interop_test/contract', 'CallScriptHash.py')
        call_contract_path = self.get_contract_path('CheckWitness.py')
        account = to_script_hash(b'NiNmXL8FjEUEs1nfX9uHFBNaenxDHJtmuB')

        engine = TestEngine()
        contract_method = 'Main'
        contract_args = [account]
        result = self.run_smart_contract(engine, call_contract_path, contract_method, *contract_args)
        contract_hash = engine.executed_script_hash.to_array()
        self.assertEqual(False, result)

        engine.add_signer_account(account)
        result = self.run_smart_contract(engine, call_contract_path, contract_method, *contract_args)
        self.assertEqual(True, result)

        engine.add_signer_account(account)
        result = self.run_smart_contract(engine, path, 'Main',
                                         contract_hash, contract_method, contract_args)
        self.assertEqual(False, result)  # fail because the signer have CalledByEntry scope

        from boa3_test.tests.test_classes.witnessscope import WitnessScope
        engine.add_signer_account(account, WitnessScope.Global)
        result = self.run_smart_contract(engine, path, 'Main',
                                         contract_hash, contract_method, contract_args)
        self.assertEqual(True, result)

    def test_check_witness_imported_as(self):
        path = self.get_contract_path('CheckWitnessImportedAs.py')
        account = to_script_hash(b'NiNmXL8FjEUEs1nfX9uHFBNaenxDHJtmuB')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', account)
        self.assertEqual(False, result)

        engine.add_signer_account(account)
        result = self.run_smart_contract(engine, path, 'Main', account)
        self.assertEqual(True, result)

    def test_check_witness_mismatched_type(self):
        path = self.get_contract_path('CheckWitnessMismatchedType.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_notify_str(self):
        event_name = String('notify').to_bytes()
        message = 'str'
        string = String(message).to_bytes()
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.PUSHDATA1
            + Integer(len(string)).to_byte_array(min_length=1)
            + string
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.SWAP
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('NotifyStr.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=Interop.Notify.name)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual((message,), event_notifications[0].arguments)

    def test_notify_int(self):
        event_name = String('notify').to_bytes()
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.PUSH15
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.SWAP
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('NotifyInt.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=Interop.Notify.name)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual((15,), event_notifications[0].arguments)

    def test_notify_bool(self):
        event_name = String('notify').to_bytes()
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.PUSHT
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.SWAP
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('NotifyBool.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=Interop.Notify.name)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual((True,), event_notifications[0].arguments)

    def test_notify_none(self):
        event_name = String('notify').to_bytes()
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.PUSHNULL
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.SWAP
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('NotifyNone.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=Interop.Notify.name)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual((None,), event_notifications[0].arguments)

    def test_notify_sequence(self):
        event_name = String('notify').to_bytes()
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(event_name)).to_byte_array(min_length=1)
            + event_name
            + Opcode.PUSH7
            + Opcode.PUSH5
            + Opcode.PUSH3
            + Opcode.PUSH2
            + Opcode.PUSH4
            + Opcode.PACK
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.SWAP
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('NotifySequence.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name=Interop.Notify.name)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual(([2, 3, 5, 7],), event_notifications[0].arguments)

    def test_notify_with_name(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x00\x01'
            + Opcode.LDARG0
            + Opcode.PUSH10
            + Opcode.PUSH1
            + Opcode.PACK
            + Opcode.SWAP
            + Opcode.SYSCALL
            + Interop.Notify.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('NotifyWithName.py')
        output, manifest = self.compile_and_save(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main', 'unit_test')
        self.assertIsVoid(result)
        self.assertGreater(len(engine.notifications), 0)

        event_notifications = engine.get_events(event_name='unit_test')
        self.assertEqual(1, len(event_notifications))
        self.assertEqual((10,), event_notifications[0].arguments)

    def test_log_mismatched_type(self):
        path = self.get_contract_path('LogMismatchedValueInt.py')
        self.assertCompilerLogs(CompilerError.MismatchedTypes, path)

    def test_log_str(self):
        string = String('str').to_bytes()
        expected_output = (
            Opcode.PUSHDATA1
            + Integer(len(string)).to_byte_array(min_length=1)
            + string
            + Opcode.SYSCALL
            + Interop.Log.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('LogStr.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertIsVoid(result)

    def test_get_trigger(self):
        expected_output = (
            Opcode.SYSCALL
            + Interop.GetTrigger.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('Trigger.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(TriggerType.APPLICATION, result)

    def test_is_application_trigger(self):
        application = Integer(TriggerType.APPLICATION.value).to_byte_array()
        expected_output = (
            Opcode.SYSCALL
            + Interop.GetTrigger.interop_method_hash
            + Opcode.PUSHINT8 + application
            + Opcode.NUMEQUAL
            + Opcode.RET
        )

        path = self.get_contract_path('TriggerApplication.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(True, result)

    def test_is_verification_trigger(self):
        verification = Integer(TriggerType.VERIFICATION.value).to_byte_array()
        expected_output = (
            Opcode.SYSCALL
            + Interop.GetTrigger.interop_method_hash
            + Opcode.PUSHINT8 + verification
            + Opcode.NUMEQUAL
            + Opcode.RET
        )

        path = self.get_contract_path('TriggerVerification.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(False, result)

    def test_get_calling_script_hash(self):
        expected_output = (
            Opcode.SYSCALL
            + Interop.CallingScriptHash.getter.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('CallingScriptHash.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        calling_hash = bytes(range(20))
        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'Main',
                                         calling_script_hash=calling_hash,
                                         expected_result_type=bytes)
        self.assertEqual(calling_hash, result)

    def test_calling_script_hash_cant_assign(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01\x01'
            + Opcode.LDARG0
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('CallingScriptHashCantAssign.py')
        output = self.assertCompilerLogs(CompilerWarning.NameShadowing, path)
        self.assertEqual(expected_output, output)

    def test_get_executing_script_hash(self):
        expected_output = (
            Opcode.SYSCALL
            + Interop.ExecutingScriptHash.getter.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('ExecutingScriptHash.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_executing_script_hash_cant_assign(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01\x01'
            + Opcode.LDARG0
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('ExecutingScriptHashCantAssign.py')
        output = self.assertCompilerLogs(CompilerWarning.NameShadowing, path)
        self.assertEqual(expected_output, output)

    def test_get_executing_script_hash_on_deploy(self):
        path = self.get_contract_path('ExecutingScriptHashOnDeploy.py')

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'get_script')
        self.assertEqual(engine.executed_script_hash.to_array(), result)

    def test_get_block_time(self):
        expected_output = (
            Opcode.SYSCALL
            + Interop.BlockTime.getter.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('BlockTime.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        new_block = engine.increase_block()
        result = self.run_smart_contract(engine, path, 'Main')
        self.assertEqual(new_block.timestamp, result)

    def test_block_time_cant_assign(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01\x01'
            + Opcode.LDARG0
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('BlockTimeCantAssign.py')
        output = self.assertCompilerLogs(CompilerWarning.NameShadowing, path)
        self.assertEqual(expected_output, output)

    def test_get_gas_left(self):
        expected_output = (
            Opcode.SYSCALL
            + Interop.GasLeft.getter.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('GasLeft.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_gas_left_cant_assign(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01\x01'
            + Opcode.LDARG0
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('GasLeftCantAssign.py')
        output = self.assertCompilerLogs(CompilerWarning.NameShadowing, path)
        self.assertEqual(expected_output, output)

    def test_get_invocation_counter(self):
        expected_output = (
            Opcode.SYSCALL
            + Interop.InvocationCounter.getter.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('InvocationCounter.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_invocation_counter_cant_assign(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01\x01'
            + Opcode.LDARG0
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('InvocationCounterCantAssign.py')
        output = self.assertCompilerLogs(CompilerWarning.NameShadowing, path)
        self.assertEqual(expected_output, output)

    def test_get_notifications(self):
        path = self.get_contract_path('GetNotifications.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'without_param', [])
        self.assertEqual(1, len(result))

        self.assertEqual(3, len(result[0]))
        # the Deploy parameter should have been the smart contract address, but the deploy method does uses a
        # ReferenceCounter and the test engine didn't replicate this behavior
        # new VM.Types.Array(engine.ReferenceCounter) { contract.Hash.ToArray() }
        event_script, event_name = result[0][:2]
        self.assertEqual(constants.MANAGEMENT_SCRIPT, event_script)
        self.assertEqual('Deploy', event_name)
        script = engine.executed_script_hash.to_array()

        engine.reset_engine()
        result = self.run_smart_contract(engine, path, 'without_param', [1, 2, 3])
        expected_result = [
            [constants.MANAGEMENT_SCRIPT, 'Deploy', script]
        ]
        for x in [1, 2, 3]:
            expected_result.append([script, 'notify', [x]])

        self.assertEqual(expected_result[1:], result[1:])

        # it's the same Deploy error
        self.assertEqual(expected_result[0][:2], result[0][:2])

        engine.reset_engine()
        result = self.run_smart_contract(engine, path, 'with_param', [], script)
        self.assertEqual([], result)

        engine.reset_engine()
        result = self.run_smart_contract(engine, path, 'with_param', [1, 2, 3], script)
        expected_result = []
        for x in [1, 2, 3]:
            expected_result.append([script,
                                    'notify',
                                    [x]])
        self.assertEqual(expected_result, result)

        engine.reset_engine()
        result = self.run_smart_contract(engine, path, 'with_param', [1, 2, 3], b'\x01' * 20)
        self.assertEqual([], result)

    def test_get_entry_script_hash(self):
        expected_output = (
            Opcode.SYSCALL
            + Interop.EntryScriptHash.getter.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('EntryScriptHash.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

    def test_entry_script_hash_cant_assign(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01\x01'
            + Opcode.LDARG0
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('EntryScriptHashCantAssign.py')
        output = self.assertCompilerLogs(CompilerWarning.NameShadowing, path)
        self.assertEqual(expected_output, output)

    def test_platform(self):
        expected_output = (
            Opcode.SYSCALL
            + Interop.Platform.getter.interop_method_hash
            + Opcode.RET
        )

        path = self.get_contract_path('Platform.py')
        output = Boa3.compile(path)
        self.assertEqual(expected_output, output)

        engine = TestEngine()
        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual('NEO', result)

    def test_platform_cant_assign(self):
        expected_output = (
            Opcode.INITSLOT
            + b'\x01\x01'
            + Opcode.LDARG0
            + Opcode.STLOC0
            + Opcode.LDLOC0
            + Opcode.RET
        )

        path = self.get_contract_path('PlatformCantAssign.py')
        output = self.assertCompilerLogs(CompilerWarning.NameShadowing, path)
        self.assertEqual(expected_output, output)

    def test_burn_gas(self):
        path = self.get_contract_path('BurnGas.py')
        engine = TestEngine()

        burn_gas_cost = 2460  # 2460 * 10^-8 GAS

        # can not burn negative GAS
        with self.assertRaisesRegex(TestExecutionException, self.GAS_MUST_BE_POSITIVE_MSG):
            self.run_smart_contract(engine, path, 'main', -10 ** 8)
        self.assertEqual(engine.gas_consumed - burn_gas_cost, 0)

        # can not burn no GAS
        with self.assertRaisesRegex(TestExecutionException, self.GAS_MUST_BE_POSITIVE_MSG):
            self.run_smart_contract(engine, path, 'main', 0)
        self.assertEqual(engine.gas_consumed - burn_gas_cost, 0)

        burned_gas = 1 * 10 ** 8  # 1 GAS
        result = self.run_smart_contract(engine, path, 'main', burned_gas)
        self.assertIsVoid(result)
        self.assertEqual(engine.gas_consumed - burn_gas_cost, burned_gas)

        burned_gas = 123 * 10 ** 5  # 0.123 GAS
        result = self.run_smart_contract(engine, path, 'main', burned_gas)
        self.assertIsVoid(result)
        self.assertEqual(engine.gas_consumed - burn_gas_cost, burned_gas)

    def test_boa2_runtime_test(self):
        path = self.get_contract_path('RuntimeBoa2Test.py')
        engine = TestEngine()

        new_block = engine.increase_block()
        result = self.run_smart_contract(engine, path, 'main', 'time', 1)
        self.assertEqual(new_block.timestamp, result)

        from boa3.builtin.type import UInt160
        result = self.run_smart_contract(engine, path, 'main', 'check_witness', UInt160(bytes(20)))
        self.assertEqual(False, result)

        result = self.run_smart_contract(engine, path, 'main', 'log', 'hello')
        self.assertEqual(True, result)

        result = self.run_smart_contract(engine, path, 'main', 'notify', 1234)
        self.assertEqual(True, result)
        event_notifications = engine.get_events(event_name=Interop.Notify.name)
        self.assertEqual(1, len(event_notifications))
        self.assertEqual((1234,), event_notifications[0].arguments)

        result = self.run_smart_contract(engine, path, 'main', 'get_trigger', 1234)
        self.assertEqual(TriggerType.APPLICATION, result)

    def test_boa2_trigger_type_test(self):
        path = self.get_contract_path('TriggerTypeBoa2Test.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main', 1)
        self.assertEqual(0x40, result)

        result = self.run_smart_contract(engine, path, 'main', 2)
        self.assertEqual(0x20, result)

        result = self.run_smart_contract(engine, path, 'main', 3)
        if isinstance(result, str):
            result = String(result).to_bytes()
        self.assertEqual(b'\x20', result)

        result = self.run_smart_contract(engine, path, 'main', 0)
        self.assertEqual(-1, result)

    def test_get_script_container(self):
        path = self.get_contract_path('ScriptContainer.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertIsNotNone(result)

    def test_get_script_container_as_transaction(self):
        path = self.get_contract_path('ScriptContainerAsTransaction.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(8, len(result))
        if isinstance(result[0], str):
            result[0] = String(result[0]).to_bytes()
        self.assertIsInstance(result[0], bytes)
        self.assertIsInstance(result[1], int)
        self.assertIsInstance(result[2], int)
        if isinstance(result[3], str):
            result[3] = String(result[3]).to_bytes()
        self.assertIsInstance(result[3], bytes)
        self.assertIsInstance(result[4], int)
        self.assertIsInstance(result[5], int)
        self.assertIsInstance(result[6], int)
        if isinstance(result[6], str):
            result[6] = String(result[6]).to_bytes()
        self.assertIsInstance(result[7], bytes)

    def test_get_network(self):
        path = self.get_contract_path('GetNetwork.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(0x334F454E, result)  # Neo3 main net's magic number

    def test_get_network_too_many_parameters(self):
        path = self.get_contract_path('GetNetworkTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_import_runtime(self):
        path = self.get_contract_path('ImportRuntime.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertIsInstance(result, int)

    def test_import_interop_runtime(self):
        path = self.get_contract_path('ImportInteropRuntime.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertIsInstance(result, int)

    def test_get_random(self):
        path = self.get_contract_path('GetRandom.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertIsInstance(result, int)

    def test_get_random_too_many_parameters(self):
        path = self.get_contract_path('GetRandomTooManyArguments.py')
        self.assertCompilerLogs(CompilerError.UnexpectedArgument, path)

    def test_address_version(self):
        path = self.get_contract_path('AddressVersion.py')
        engine = TestEngine()

        result = self.run_smart_contract(engine, path, 'main')
        self.assertEqual(53, result)    # current Neo protocol version is 53

    def test_address_version_cant_assign(self):
        path = self.get_contract_path('AddressVersionCantAssign.py')
        self.assertCompilerLogs(CompilerWarning.NameShadowing, path)

    def test_load_script(self):
        path = self.get_contract_path('LoadScriptDynamicCall.py')
        self.compile_and_save(path)
        engine = TestEngine()

        operand_1 = 1
        operand_2 = 2
        expected_result = operand_1 + operand_2
        result = self.run_smart_contract(engine, path, 'dynamic_sum',
                                         operand_1, operand_2)
        self.assertEqual(expected_result, result)

        from boa3.neo3.contracts import CallFlags
        result = self.run_smart_contract(engine, path, 'dynamic_sum_with_flags',
                                         operand_1, operand_2, CallFlags.READ_ONLY)
        self.assertEqual(expected_result, result)
